from tooldelta.basic_mods import dotcs_module_env
from tooldelta.color_print import Print
from tooldelta.get_python_libs import get_single_lib
from tooldelta.plugin_load import (
    NON_FUNC,
    classic_plugin,
    dotcs_plugin,
    injected_plugin,
)
from tooldelta.plugin_load.classic_plugin import Plugin, PluginAPI


import asyncio
import importlib
import threading
import time
import traceback
from typing import Any, Callable, Type


class PluginGroup:
    class PluginAPINotFoundError(ModuleNotFoundError):
        def __init__(self, name):
            self.name = name

    class PluginAPIVersionError(ModuleNotFoundError):
        def __init__(self, name, m_ver, n_ver):
            self.name = name
            self.m_ver = m_ver
            self.n_ver = n_ver

    plugins: list[Plugin] = []
    plugins_funcs: dict[str, list] = {
        "on_def": [],
        "on_inject": [],
        "on_player_prejoin": [],
        "on_player_join": [],
        "on_player_message": [],
        "on_player_death": [],
        "on_player_leave": [],
    }
    plugin_added_cache = {"plugin": None, "packets": []}
    pluginAPI_added_cache = []

    def __init__(self, frame, PRG_NAME):
        self.listen_packet_ids = set()
        self.old_dotcs_env = {}
        self.dotcs_global_vars = {}
        self.packet_funcs: dict[str, list[Callable]] = {}
        self.plugins_api: dict[str, PluginAPI] = {}
        self.excType = 0
        self.PRG_NAME = ""
        self._broadcast_evts = {}
        self.dotcs_plugin_loaded_num = 0
        self.normal_plugin_loaded_num = 0
        self.injected_plugin_loaded_num = 0
        self.linked_frame = frame
        self.PRG_NAME = PRG_NAME
        self.dotcs_repeat_threadings = {"1s": [], "10s": [], "30s": [], "1m": []}
        self.linked_frame.linked_plugin_group = self

    @staticmethod
    def require(module_name: str, pip_name=""):
        try:
            importlib.import_module(module_name)
        except (ModuleNotFoundError, ImportError):
            get_single_lib(pip_name if pip_name else module_name)

    def read_all_plugins(self):
        dotcs_plugin.read_plugin_from_old(self, dotcs_module_env)
        classic_plugin.read_plugin_from_new(self, {})
        asyncio.run(injected_plugin.load_plugin(self))

    def add_broadcast_listener(self, evt_name: str):
        "将下面的方法作为一个广播事件接收器"

        def deco(func: Callable[[Any], bool]):
            if self._broadcast_evts.get(evt_name):
                self._broadcast_evts[evt_name].append(func)
            else:
                self._broadcast_evts[evt_name] = [func]

        return deco

    def broadcastEvt(self, evt_name: str, **kwargs) -> list[Any] | None:
        "向全局广播一个特定事件, 可以传入附加信息参数"
        callback_list = []
        res = self._broadcast_evts.get(evt_name)
        if res:
            for f in res:
                interrupt, *res2 = f(**kwargs)
                if res2:
                    callback_list.append(res2)
                    if interrupt:
                        break
            return callback_list
        return None

    def add_plugin(self, plugin):
        try:
            if not Plugin.__subclasscheck__(plugin):
                raise AssertionError(1, f"插件主类必须继承Plugin类 而不是 {plugin}")
        except TypeError:
            if not Plugin.__subclasscheck__(type(plugin)):
                raise AssertionError(
                    1, f"插件主类必须继承Plugin类 而不是 {plugin.__class__.__name__}"
                )
        self.plugin_added_cache["plugin"] = plugin
        return plugin

    def add_packet_listener(self, pktID):
        def deco(func):
            self.plugin_added_cache["packets"].append((pktID, func))
            return func

        return deco

    def add_plugin_api(self, apiName: str):
        def _add_api(api: Type[PluginAPI]):
            if not PluginAPI.__subclasscheck__(api):
                raise AssertionError(1, "插件API类必须继承PluginAPI类")
            self.pluginAPI_added_cache.append((apiName, api))

        return _add_api

    def add_plugin_as_api(self, apiName: str):
        def _add_plugin_2_api(api_plugin: Type[PluginAPI]):
            if not PluginAPI.__subclasscheck__(api_plugin):
                raise AssertionError(
                    1,
                    "API插件API类必须继承PluginAPI类和Plugin类",
                )
            self.plugin_added_cache["plugin"] = api_plugin
            self.pluginAPI_added_cache.append(apiName)

        return _add_plugin_2_api

    def get_plugin_api(
        self, apiName: str, min_version: tuple | None = None
    ) -> PluginAPI:
        api = self.plugins_api.get(apiName)
        if api:
            if min_version and api.version < min_version:
                raise self.PluginAPIVersionError(apiName, min_version, api.version)
            return api
        raise self.PluginAPINotFoundError(apiName)

    def checkSystemVersion(self, need_vers: tuple[int, int, int]):
        if need_vers > self.linked_frame.sys_data.system_version:
            raise self.linked_frame.SystemVersionException(
                f"该组件需要{self.linked_frame.PRG_NAME}为{'.'.join([str(i) for i in self.linked_frame.sys_data.system_version])}版本"
            )

    def add_listen_packet_id(self, packetType: int):
        self.listen_packet_ids.add(packetType)
        self.linked_frame.link_game_ctrl.add_listen_pkt(packetType)

    def add_listen_packet_func(self, packetType, func: Callable):
        if self.packet_funcs.get(str(packetType)):
            self.packet_funcs[str(packetType)].append(func)
        else:
            self.packet_funcs[str(packetType)] = [func]

    def execute_dotcs_repeat(self):
        "启动dotcs插件的循环执行模式插件事件"
        threading.Thread(target=self.run_dotcs_repeat_funcs).start()

    def run_dotcs_repeat_funcs(self):
        lastTime10s = 0
        lastTime30s = 0
        lastTime1m = 0
        if not any(self.dotcs_repeat_threadings.values()):
            return
        Print.print_inf(
            f"开始运行 {sum([len(funcs) for funcs in self.dotcs_repeat_threadings.values()])} 个原dotcs计划任务方法"
        )
        while 1:
            time.sleep(1)
            nowTime = time.time()
            if nowTime - lastTime1m > 60:
                for fname, func in self.dotcs_repeat_threadings["1m"]:
                    try:
                        # A strong desire to remove "try" block !!
                        func()
                    except Exception as err:
                        Print.print_err(f"原dotcs插件 <{fname}> (计划任务1min)报错: {err}")
                lastTime1m = nowTime
            if nowTime - lastTime30s > 30:
                for fname, func in self.dotcs_repeat_threadings["30s"]:
                    try:
                        func()
                    except Exception as err:
                        Print.print_err(f"原dotcs插件 <{fname}> (计划任务30s)报错: {err}")
                lastTime30s = nowTime
            if nowTime - lastTime10s > 10:
                for fname, func in self.dotcs_repeat_threadings["10s"]:
                    try:
                        func()
                    except Exception as err:
                        Print.print_err(f"原dotcs插件 <{fname}> (计划任务10s)报错: {err}")
                lastTime10s = nowTime
            for fname, func in self.dotcs_repeat_threadings["1s"]:
                try:
                    func()
                except Exception as err:
                    Print.print_err(f"原dotcs插件 <{fname}> (计划任务1s) 报错: {err}")

    def execute_def(self, onerr: Callable[[str, Exception, str], None] = NON_FUNC):
        for name, func in self.plugins_funcs["on_def"]:
            try:
                func()
            except PluginGroup.PluginAPINotFoundError as err:
                Print.print_err(f"插件 {name} 需要包含该种接口的前置组件: {err.name}")
                raise SystemExit
            except PluginGroup.PluginAPIVersionError as err:
                Print.print_err(
                    f"插件 {name} 需要该前置组件 {err.name} 版本: {err.m_ver}, 但是现有版本过低: {err.n_ver}"
                )
                raise SystemExit
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_init(self, onerr: Callable[[str, Exception, str], None] = NON_FUNC):
        for name, func in self.plugins_funcs["on_inject"]:
            try:
                func()
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_prejoin(
        self, player, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        for name, func in self.plugins_funcs["on_player_prejoin"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_join(
        self, player, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        for name, func in self.plugins_funcs["on_player_join"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_message(
        self, player, msg, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        pat = f"[{player}] "
        if msg.startswith(pat):
            msg = msg.strip(pat)
        for name, func in self.plugins_funcs["on_player_message"]:
            try:
                func(player, msg)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_leave(
        self, player, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        for name, func in self.plugins_funcs["on_player_leave"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_death(
        self,
        player: str,
        killer: str | None,
        msg: str,
        onerr: Callable[[str, Exception, str], None] = NON_FUNC,
    ):
        for name, func in self.plugins_funcs["on_player_death"]:
            try:
                func(player, killer, msg)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def processPacketFunc(self, pktID: int, pkt: dict):
        d = self.packet_funcs.get(str(pktID))
        if d:
            for func in d:
                try:
                    res = func(pkt)
                    if res:
                        return True
                except:
                    Print.print_err(f"插件方法 {func.__name__} 出错：")
                    Print.print_err(traceback.format_exc())
        return False



# Manage libraries
try:
    import time
    import os
    import re
    import sys
    import threading
    import json
    import traceback
    import datetime
    import platform
    import subprocess
    import socket
    import logging
    import ctypes
    import asyncio
    import copy
    import math
    import random
    import psutil
    import requests
    import nbt
    import qrcode
    import getpass
    import pymysql
    import websockets
    import ping3
    import ujson
    import hashlib
    import base64
    import rich

    sys.path.append(os.path.join(os.getcwd(), "libs"))

except ModuleNotFoundError as err:
    # 第一次部署该项目, 将会自动安装这些模块.
    from . import get_python_libs

    get_python_libs.try_install_libs(err)
    import psutil
    import requests
    import nbt
    import qrcode
    import getpass
    import pymysql
    import websockets
    import ujson
    import hashlib
    import base64
    import rich

from typing import Callable

dotcs_module_env = {
    "time": time,
    "os": os,
    "json": json,
    "threading": threading,
    "requests": requests,
    "traceback": traceback,
    "qrcode": qrcode,
    "psutil": psutil,
    "pymysql": pymysql,
    "socket": socket,
    "websockets": websockets,
    "datetime": datetime,
    "random": random,
    "math": math,
    "platform": platform,
    "subprocess": subprocess,
    "base64": base64,
    "asyncio": asyncio,
    "sys": sys,
    "rich": rich,
}



class PacketIDS:
    Text = 9
    PlayerList = 63
    CommandOutput = 79


packets = {
    "1": "IDLogin",
    "2": "IDPlayStatus",
    "3": "IDServerToClientHandshake",
    "4": "IDClientToServerHandshake",
    "5": "IDDisconnect",
    "6": "IDResourcePacksInfo",
    "7": "IDResourcePackStack",
    "8": "IDResourcePackClientResponse",
    "9": "IDText",
    "10": "IDSetTime",
    "11": "IDStartGame",
    "12": "IDAddPlayer",
    "13": "IDAddActor",
    "14": "_",
    "15": "IDRemoveActor",
    "16": "IDAddItemActor",
    "17": "IDTakeItemActor",
    "18": "IDMoveActorAbsolute",
    "19": "IDMovePlayer",
    "20": "IDPassengerJump",
    "21": "IDUpdateBlock",
    "22": "IDAddPainting",
    "23": "IDTickSync",
    "24": "_",
    "25": "IDLevelEvent",
    "26": "IDBlockEvent",
    "27": "IDActorEvent",
    "28": "IDMobEffect",
    "29": "IDUpdateAttributes",
    "30": "IDInventoryTransaction",
    "31": "IDMobEquipment",
    "32": "IDMobArmourEquipment",
    "33": "IDInteract",
    "34": "IDBlockPickRequest",
    "35": "IDActorPickRequest",
    "36": "IDPlayerAction",
    "37": "_",
    "38": "IDHurtArmour",
    "39": "IDSetActorData",
    "40": "IDSetActorMotion",
    "41": "IDSetActorLink",
    "42": "IDSetHealth",
    "43": "IDSetSpawnPosition",
    "44": "IDAnimate",
    "45": "IDRespawn",
    "46": "IDContainerOpen",
    "47": "IDContainerClose",
    "48": "IDPlayerHotBar",
    "49": "IDInventoryContent",
    "50": "IDInventorySlot",
    "51": "IDContainerSetData",
    "52": "IDCraftingData",
    "53": "IDCraftingEvent",
    "54": "IDGUIDataPickItem",
    "55": "IDAdventureSettings",
    "56": "IDBlockActorData",
    "57": "IDPlayerInput",
    "58": "IDLevelChunk",
    "59": "IDSetCommandsEnabled",
    "60": "IDSetDifficulty",
    "61": "IDChangeDimension",
    "62": "IDSetPlayerGameType",
    "63": "IDPlayerList",
    "64": "IDSimpleEvent",
    "65": "IDEvent",
    "66": "IDSpawnExperienceOrb",
    "67": "IDClientBoundMapItemData",
    "68": "IDMapInfoRequest",
    "69": "IDRequestChunkRadius",
    "70": "IDChunkRadiusUpdated",
    "71": "IDItemFrameDropItem",
    "72": "IDGameRulesChanged",
    "73": "IDCamera",
    "74": "IDBossEvent",
    "75": "IDShowCredits",
    "76": "IDAvailableCommands",
    "77": "IDCommandRequest",
    "78": "IDCommandBlockUpdate",
    "79": "IDCommandOutput",
    "80": "IDUpdateTrade",
    "81": "IDUpdateEquip",
    "82": "IDResourcePackDataInfo",
    "83": "IDResourcePackChunkData",
    "84": "IDResourcePackChunkRequest",
    "85": "IDTransfer",
    "86": "IDPlaySound",
    "87": "IDStopSound",
    "88": "IDSetTitle",
    "89": "IDAddBehaviourTree",
    "90": "IDStructureBlockUpdate",
    "91": "IDShowStoreOffer",
    "92": "IDPurchaseReceipt",
    "93": "IDPlayerSkin",
    "94": "IDSubClientLogin",
    "95": "IDAutomationClientConnect",
    "96": "IDSetLastHurtBy",
    "97": "IDBookEdit",
    "98": "IDNPCRequest",
    "99": "IDPhotoTransfer",
    "100": "IDModalFormRequest",
    "101": "IDModalFormResponse",
    "102": "IDServerSettingsRequest",
    "103": "IDServerSettingsResponse",
    "104": "IDShowProfile",
    "105": "IDSetDefaultGameType",
    "106": "IDRemoveObjective",
    "107": "IDSetDisplayObjective",
    "108": "IDSetScore",
    "109": "IDLabTable",
    "110": "IDUpdateBlockSynced",
    "111": "IDMoveActorDelta",
    "112": "IDSetScoreboardIdentity",
    "113": "IDSetLocalPlayerAsInitialised",
    "114": "IDUpdateSoftEnum",
    "115": "IDNetworkStackLatency",
    "116": "_",
    "117": "IDScriptCustomEvent",
    "118": "IDSpawnParticleEffect",
    "119": "IDAvailableActorIdentifiers",
    "120": "_",
    "121": "IDNetworkChunkPublisherUpdate",
    "122": "IDBiomeDefinitionList",
    "123": "IDLevelSoundEvent",
    "124": "IDLevelEventGeneric",
    "125": "IDLecternUpdate",
    "126": "_",
    "127": "IDAddEntity",
    "128": "IDRemoveEntity",
    "129": "IDClientCacheStatus",
    "130": "IDMapCreateLockedCopy",
    "131": "IDOnScreenTextureAnimation",
    "132": "IDStructureTemplateDataRequest",
    "133": "IDStructureTemplateDataResponse",
    "134": "_",
    "135": "IDClientCacheBlobStatus",
    "136": "IDClientCacheMissResponse",
    "137": "IDEducationSettings",
    "138": "IDEmote",
    "139": "IDMultiPlayerSettings",
    "140": "IDSettingsCommand",
    "141": "IDAnvilDamage",
    "142": "IDCompletedUsingItem",
    "143": "IDNetworkSettings",
    "144": "IDPlayerAuthInput",
    "145": "IDCreativeContent",
    "146": "IDPlayerEnchantOptions",
    "147": "IDItemStackRequest",
    "148": "IDItemStackResponse",
    "149": "IDPlayerArmourDamage",
    "150": "IDCodeBuilder",
    "151": "IDUpdatePlayerGameType",
    "152": "IDEmoteList",
    "153": "IDPositionTrackingDBServerBroadcast",
    "154": "IDPositionTrackingDBClientRequest",
    "155": "IDDebugInfo",
    "156": "IDPacketViolationWarning",
    "157": "IDMotionPredictionHints",
    "158": "IDAnimateEntity",
    "159": "IDCameraShake",
    "160": "IDPlayerFog",
    "161": "IDCorrectPlayerMovePrediction",
    "162": "IDItemComponent",
    "163": "IDFilterText",
    "164": "IDClientBoundDebugRenderer",
    "165": "IDSyncActorProperty",
    "166": "IDAddVolumeEntity",
    "167": "IDRemoveVolumeEntity",
    "168": "IDSimulationType",
    "169": "IDNPCDialogue",
    "170": "IDEducationResourceURI",
    "171": "IDCreatePhoto",
    "172": "IDUpdateSubChunkBlocks",
    "173": "IDPhotoInfoRequest",
    "174": "IDSubChunk",
    "175": "IDSubChunkRequest",
    "176": "IDClientStartItemCooldown",
    "177": "IDScriptMessage",
    "178": "IDCodeBuilderSource",
}


class SubPacket_CmdOutputMsg:
    Success: bool
    Message: str
    Parameters: list[str]

    def __init__(self, pkt: dict):
        self.Success = pkt["Success"]
        self.Parameters = pkt["Parameters"]
        self.Message = pkt["Message"]


class SubPacket_CmdOrigin:
    Origin: int
    UUID: str
    RequestID: str
    PlayerUniqueID: int

    def __init__(self, pkt: dict):
        self.Origin = pkt["Origin"]
        self.UUID = pkt["UUID"]
        self.RequestID = pkt["RequestID"]
        self.PlayerUniqueID = pkt["PlayerUniqueID"]


class Packet_CommandOutput:
    CommandOrigin: SubPacket_CmdOrigin
    OutputType: int
    SuccessCount: int
    OutputMessages: list[SubPacket_CmdOutputMsg]
    as_dict: dict

    def __init__(self, pkt: dict):
        self.as_dict = pkt
        self.CommandOrigin = SubPacket_CmdOrigin(pkt["CommandOrigin"])
        self.OutputMessages = [
            SubPacket_CmdOutputMsg(imsg) for imsg in pkt["OutputMessages"]
        ]
        self.SuccessCount = pkt["SuccessCount"]
        self.OutputType = pkt["OutputType"]



from .color_print import Print
import ujson
import os
import threading
import traceback
import copy
import ctypes
from typing import Any, Dict

event_pool = {"tmpjson_save": threading.Event()}
event_flags_pool = {"tmpjson_save": True}


class Builtins:
    class ThreadExit(SystemExit):
        "线程退出."

    class ClassicThread(threading.Thread):
        def __init__(self, func, args: tuple = (), usage="", **kwargs):
            """
            方便地新建一个ToolDelta子线程

            参数:
                func: 线程方法
                args: 方法的参数项 tuple
                usage: 线程的用途说明
                ..kwargs: 方法的关键字参数项
            """
            super().__init__(target=func)
            self.func = func
            self.daemon = True
            self.all_args = [args, kwargs]
            self.usage = usage
            self.start()

        def run(self):
            try:
                self.func(*self.all_args[0], **self.all_args[1])
            except Builtins.ThreadExit:
                pass
            except:
                Print.print_err(f"线程 {self.usage} 出错:\n" + traceback.format_exc())

    createThread = ClassicThread

    class TMPJson:
        @staticmethod
        def loadPathJson(path: str, needFileExists: bool = True):
            """
            将json文件从磁盘加载到缓存区, 以便快速读写.
            在缓存文件已加载的情况下, 再使用一次该方法不会有任何作用.

            参数:
                path: 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
                needFileExists: 默认为 True, 为 False 时, 若文件路径不存在, 就会自动创建一个文件.
            """
            if path in jsonPathTmp:
                return
            try:
                js = Builtins.SimpleJsonDataReader.SafeJsonLoad(path)
            except FileNotFoundError as err:
                if not needFileExists:
                    js = None
                else:
                    raise err from None
            jsonPathTmp[path] = [False, js]

        @staticmethod
        def unloadPathJson(path: str):
            """
            将json文件从缓存区卸载(保存内容到磁盘), 之后不能再在缓存区对这个文件进行读写.
            在缓存文件已卸载的情况下, 再使用一次该方法不会有任何作用, 但是可以通过其返回的值来知道存盘有没有成功.

            参数:
                path: 文件的虚拟路径
            """
            if jsonPathTmp.get(path) is not None:
                isChanged, dat = jsonPathTmp[path]
                if isChanged:
                    Builtins.SimpleJsonDataReader.SafeJsonDump(dat, path)
                del jsonPathTmp[path]
                return True
            return False

        @staticmethod
        def read(path: str):
            """
            对缓存区的该虚拟路径的文件进行读操作

            参数:
                path: 文件的虚拟路径
            """
            if path in jsonPathTmp:
                val = jsonPathTmp.get(path)[1]
                if isinstance(val, (list, dict)):
                    val = copy.deepcopy(val)
                return val
            raise Exception("json路径未初始化, 不能进行读取和写入操作: " + path)

        @staticmethod
        def write(path: str, obj: Any):
            """
            对缓存区的该虚拟路径的文件进行写操作, 这将会覆盖之前的内容

            参数:
                path: 文件的虚拟路径
                obj: 任何合法的JSON类型 例如 dict/list/str/bool/int/float
            """
            if path in jsonPathTmp:
                jsonPathTmp[path] = [True, obj]
            else:
                raise Exception("json路径未初始化, 不能进行读取和写入操作: " + path)

        @staticmethod
        def cancel_change(path):
            "取消缓存json所做的更改, 非必要情况请勿调用, 你不知道什么时候会自动保存所做更改"
            jsonPathTmp[path][0] = False

        @staticmethod
        def get_tmps():
            "不要调用!"
            return jsonPathTmp.copy()

    class SimpleJsonDataReader:
        @staticmethod
        def SafeJsonDump(obj: str | dict | list, fp):
            """
            导出一个json文件, 会自动关闭文件读写接口.

            参数:
                obj: json对象.
                fp: 由open(...)打开的文件读写口 或 文件路径.
            """
            if isinstance(fp, str):
                with open(fp, "w", encoding="utf-8") as file:
                    file.write(ujson.dumps(obj, indent=4, ensure_ascii=False))
            else:
                with fp:
                    fp.write(ujson.dumps(obj, indent=4, ensure_ascii=False))

        @staticmethod
        def SafeJsonLoad(fp):
            """
            读取一个json文件, 会自动关闭文件读写接口.

            参数:
                fp: open(...)打开的文件读写口 或 文件路径.
            """
            if isinstance(fp, str):
                with open(fp, "r", encoding="utf-8") as file:
                    return ujson.loads(file.read())
            with fp as file:
                return ujson.loads(file.read())

        class DataReadError(ujson.JSONDecodeError):
            ...

        @staticmethod
        def readFileFrom(plugin_name: str, file: str, default: dict = None):
            """
            使用插件便捷地读取一个json文件,
            这个文件应在 data/<plugin_name>/<file>文件夹内, 文件夹不存在时也会自动创建

            参数:
                plugin_name: 插件名
                file: 文件名
                default: 不为None时: 当文件不存在则创建一个空文件, 使用default给出的json字典写入文件.
            """
            filepath = os.path.join("data", plugin_name, f"{file}.json")
            os.makedirs(os.path.join("data", plugin_name), exist_ok=True)
            try:
                if default is not None and not os.path.isfile(filepath):
                    with open(filepath, "w", encoding="utf-8") as f:
                        Builtins.SimpleJsonDataReader.SafeJsonDump(default, f)
                    return default
                with open(filepath, "r", encoding="utf-8") as f:
                    res = Builtins.SimpleJsonDataReader.SafeJsonLoad(f)
                return res
            except ujson.JSONDecodeError as err:
                raise Builtins.SimpleJsonDataReader.DataReadError(
                    err.msg, err.doc, err.pos
                )
            except Exception as err:
                Print.print_err(f"读文件路径 {filepath} 发生错误")
                raise err

        @staticmethod
        def writeFileTo(plugin_name: str, file: str, obj):
            """
            使用插件简单地写入一个json文件, 会覆盖原有内容
            这个文件应在data/<plugin_name>/<file>文件夹内

            参数:
                plugin_name: 插件名
                file: 文件名
                obj: 任何合法的JSON类型 例如 dict/list/str/bool/int/float
            """
            os.makedirs(f"data/{plugin_name}", exist_ok=True)
            with open(f"data/{plugin_name}/{file}.json", "w", encoding="utf-8") as f:
                Builtins.SimpleJsonDataReader.SafeJsonDump(obj, f)

    @staticmethod
    def SimpleFmt(kw: dict[str, object], __sub: str):
        """
        快速将字符串内按照给出的dict的键值对替换掉内容.
        参数:
            kw: dict
            __sub: 需要被替换的字符串

        Example:
        >>> my_color = "red"; my_item = "apple"
        >>> kw = {"[颜色]": my_color, "[物品]": my_item}
        >>> SimpleFmt(kw, "I like [颜色] [物品].")
        I like red apple.
        """
        for k, v in kw.items():
            if k in __sub:
                __sub = __sub.replace(k, str(v))
        return __sub

    @staticmethod
    def simpleAssert(cond: any, exc):
        """相当于 assert cond, 但是可以自定义引发的异常的类型"""
        if not cond:
            raise exc

    @staticmethod
    def new_thread(func):
        """
        在事件方法可能执行较久会造成堵塞时使用, 方便快捷地创建一个新线程, 例如:
        @Builtins.run_as_new_thread
        def on_inject(self):
            ...
        """

        def thread_fun(*args, **kwargs):
            Builtins.createThread(func, args=args, **kwargs)

        return thread_fun

    run_as_new_thread = new_thread

    @staticmethod
    def try_int(arg):
        """
        尝试将提供的参数化为int类型并返回, 否则返回None
        """
        try:
            return int(arg)
        except:
            return None

    @staticmethod
    def add_in_dialogue_player(player: str):
        """
        使玩家进入聊天栏对话模式, 可防止其在对话时继续触发另一个会话线程

        参数:
            玩家名: str
        """
        if player not in in_dialogue_list:
            in_dialogue_list.append(player)
        else:
            raise Exception("Already in a dialogue!")

    @staticmethod
    def remove_in_dialogue_player(player: str):
        """
        使玩家离开聊天栏对话模式

        参数:
            player: 玩家名
        """
        if player not in in_dialogue_list:
            return
        in_dialogue_list.remove(player)

    @staticmethod
    def player_in_dialogue(player: str) -> bool:
        """
        检测玩家是否处在聊天栏对话模式中.

        参数:
            player: 玩家名
        返回:
            检测结果: bool
        """
        return player in in_dialogue_list

    @staticmethod
    def create_dialogue_threading(player, func, exc_cb=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        """
        创建一个玩家与聊天栏交互的线程, 
        线程启动时玩家会自动进入聊天栏对话模式
        线程结束后该玩家会自动退出聊天栏对话模式
        可以用来防止玩家多开聊天栏对话或菜单线程
        
        参数:
            player: 玩家名
            func: 线程方法
            exc_cb: 若玩家已处于一个对话中, 则向方法exc_cb传参并调用它: player(玩家名)
            args: 线程方法的参数组 tuple
            kwargs: 线程方法的关键词参数组: dict
        """
        Builtins.createThread(
            _dialogue_thread_run, args=(player, func, exc_cb, args, kwargs)
        )

    @staticmethod
    def fuzzy_match(lst: list, sub: str):
        """
        模糊匹配列表内的字符串, 可以用在诸如模糊匹配玩家名的用途
        Example:
        >>> fuzzy_match(["a", "ab", "abc", "abd"], "ab")
        ["ab", "abc", "abd"]

        参数:
            lst: 字符串列表
            sub: 需要匹配的字符串
        """
        res = []
        for i in lst:
            if sub in i:
                res.append(i)
        return res

    class ArgsReplacement:
        def __init__(self, kw: Dict[str, Any]):
            self.kw = kw

        def replaceTo(self, __sub: str):
            for k, v in self.kw.items():
                if k in __sub:
                    __sub = __sub.replace(k, str(v))
            return __sub


def safe_close():
    event_pool["tmpjson_save"].set()
    event_flags_pool["tmpjson_save"] = False


def _tmpjson_save_thread():
    evt = event_pool["tmpjson_save"]
    while 1:
        evt.wait(60)
        for k, (isChanged, dat) in jsonPathTmp.copy().items():
            if isChanged:
                Builtins.SimpleJsonDataReader.SafeJsonDump(dat, k)
                jsonPathTmp[k][0] = False
        if not event_flags_pool["tmpjson_save"]:
            return


def tmpjson_save_thread():
    Builtins.createThread(_tmpjson_save_thread)


def _dialogue_thread_run(player, func, exc_cb, args, kwargs):
    if not Builtins.player_in_dialogue(player):
        Builtins.add_in_dialogue_player(player)
    else:
        if exc_cb is not None:
            exc_cb(player)
        return
    try:
        func(*args, **kwargs)
    except:
        Print.print_err(f"玩家{player}的会话线程 出现问题:")
        Print.print_err(traceback.format_exc())
    Builtins.remove_in_dialogue_player(player)


jsonPathTmp = {}
in_dialogue_list = []



import importlib
import os
import sys
import traceback
from ...color_print import Print
from ...builtins import Builtins
from ...cfg import Cfg
from ..funcs import unzip_plugin


class Plugin:
    name = "<未命名插件>"
    version = (0, 0, 1)
    author = "?"


class PluginAPI:
    name = "<未命名插件api>"
    version = (0, 0, 1)

    def __init__(self, _):
        raise Exception("需要初始化__init__方法")


def read_plugin_from_new(plugin_grp, root_env: dict):
    PLUGIN_PATH = os.path.join(os.getcwd(), "插件文件/ToolDelta组合式插件")
    for plugin_dir in os.listdir(PLUGIN_PATH):
        if (
            not os.path.isdir(os.path.join(PLUGIN_PATH, plugin_dir.strip(".zip")))
            and os.path.isfile(os.path.join(PLUGIN_PATH, plugin_dir))
            and plugin_dir.endswith(".zip")
        ):
            Print.print_with_info(f"§6正在解压插件{plugin_dir}, 请稍后", "§6 解压 ")
            unzip_plugin(
                os.path.join(PLUGIN_PATH, plugin_dir),
                os.path.join("插件文件/ToolDelta组合式插件", plugin_dir.strip(".zip")),
            )
            Print.print_suc(f"§a成功解压插件{plugin_dir} -> 插件目录")
            plugin_dir = plugin_dir.strip(".zip")
        if os.path.isdir(os.path.join(PLUGIN_PATH, plugin_dir)):
            sys.path.append(os.path.join(PLUGIN_PATH, plugin_dir))
            plugin_grp.plugin_added_cache["plugin"] = None
            plugin_grp.plugin_added_cache["packets"].clear()
            plugin_grp.pluginAPI_added_cache.clear()
            try:
                if os.path.isfile(
                    os.path.join("插件文件/ToolDelta组合式插件", plugin_dir, "__init__.py")
                ):
                    sys.path.append(os.path.join(os.getcwd(), "插件文件", "ToolDelta组合式插件"))
                    # 理论上所有插件共享一个整体环境
                    importlib.__import__(plugin_dir, root_env)
                else:
                    Print.print_war(f"{plugin_dir} 文件夹 未发现插件文件, 跳过加载")
                    continue
                if plugin_grp.plugin_added_cache["plugin"] is None:
                    raise AssertionError(2)
                plugin = plugin_grp.plugin_added_cache["plugin"]
                plugin_body: Plugin = plugin(plugin_grp.linked_frame)
                plugin_grp.plugins.append([plugin_body.name, plugin_body])
                _v0, _v1, _v2 = plugin_body.version  # type: ignore
                if hasattr(plugin_body, "on_def"):
                    plugin_grp.plugins_funcs["on_def"].append(
                        [plugin_body.name, plugin_body.on_def]
                    )
                if hasattr(plugin_body, "on_inject"):
                    plugin_grp.plugins_funcs["on_inject"].append(
                        [plugin_body.name, plugin_body.on_inject]
                    )
                if hasattr(plugin_body, "on_player_prejoin"):
                    plugin_grp.plugins_funcs["on_player_prejoin"].append(
                        [plugin_body.name, plugin_body.on_player_prejoin]
                    )
                if hasattr(plugin_body, "on_player_join"):
                    plugin_grp.plugins_funcs["on_player_join"].append(
                        [plugin_body.name, plugin_body.on_player_join]
                    )
                if hasattr(plugin_body, "on_player_message"):
                    plugin_grp.plugins_funcs["on_player_message"].append(
                        [plugin_body.name, plugin_body.on_player_message]
                    )
                if hasattr(plugin_body, "on_player_death"):
                    plugin_grp.plugins_funcs["on_player_death"].append(
                        [plugin_body.name, plugin_body.on_player_death]
                    )
                if hasattr(plugin_body, "on_player_leave"):
                    plugin_grp.plugins_funcs["on_player_leave"].append(
                        [plugin_body.name, plugin_body.on_player_leave]
                    )
                Print.print_suc(
                    f"成功载入插件 {plugin_body.name} 版本: {_v0}.{_v1}.{_v2} 作者：{plugin_body.author}"
                )
                plugin_grp.normal_plugin_loaded_num += 1
                if plugin_grp.plugin_added_cache["packets"] != []:
                    for pktType, func in plugin_grp.plugin_added_cache["packets"]:  # type: ignore
                        plugin_grp.add_listen_packet_id(pktType)
                        plugin_grp.add_listen_packet_func(
                            pktType, getattr(plugin_body, func.__name__)
                        )
                if plugin_grp.pluginAPI_added_cache is not None:
                    for _api in plugin_grp.pluginAPI_added_cache:
                        if isinstance(_api, str):
                            plugin_grp.plugins_api[_api] = plugin_body
                        else:
                            (apiName, api) = _api
                            plugin_grp.plugins_api[apiName] = api(
                                plugin_grp.linked_frame
                            )
            except AssertionError as err:
                if err.args[0] == 2:
                    Print.print_err(
                        f"插件 {plugin_dir} 不合法: 只能调用一次 @plugins.add_plugin, 实际调用了0次或多次"
                    )
                    raise SystemExit
                elif isinstance(err.args[0], int):
                    Print.print_err(f"插件 {plugin_dir} 不合法: {err.args[1]}")
                    raise SystemExit
                if len(err.args[0]) == 2:
                    Print.print_err(f"插件 {plugin_dir} 不合法: {err.args[0][1]}")
                    raise SystemExit
                raise
            except Cfg.ConfigError as err:
                Print.print_err(f"插件 {plugin_dir} 配置文件报错：{err}")
                Print.print_err("你也可以直接删除配置文件, 重新启动ToolDelta以自动生成配置文件")
                raise SystemExit
            except Builtins.SimpleJsonDataReader.DataReadError as err:
                Print.print_err(f"插件 {plugin_dir} 读取数据失败: {err}")
            except plugin_grp.linked_frame.SystemVersionException:
                Print.print_err(f"插件 {plugin_dir} 需要更高版本的ToolDelta加载: {err}")
            except Exception as err:
                if "() takes no arguments" in str(err):
                    Print.print_err(f"插件 {plugin_dir} 不合法： 主类初始化时应接受 1 个参数: Frame")
                else:
                    Print.print_err(f"加载插件 {plugin_dir} 出现问题, 报错如下: ")
                    Print.print_err("§c" + traceback.format_exc())
                    raise SystemExit



import shlex
import requests
import time
import os
import platform
from .color_print import Print
import shutil
import socket


def _pretty_kb(n):
    if n >= 1048576:
        return f"{round(n / 1048576, 2)}M"
    if n >= 1024:
        return f"{round(n / 1024, 2)}K"
    return f"{round(n, 1)}"


def _path_get_filename(path: str):
    if "/" not in path:
        return None
    return path.split("/")[-1]


def _is_common_text_file(url_path: str):
    return any(
        url_path.endswith(i)
        for i in [".txt", ".yml", ".md", ".xml", ".py", ".h", ".c", ".pyi", ".json"]
    )


def get_file_size(url):
    response = requests.head(url)
    if "Content-Length" in response.headers:
        file_size = int(response.headers["Content-Length"])
        return file_size
    return None


def download_file(f_url: str, f_dir: str, ignore_warnings=False):
    with requests.get(f_url, stream=True, timeout=10) as res:
        res.raise_for_status()
        filesize = get_file_size(f_url)

        if filesize < 256 and not ignore_warnings:
            Print.print_war(f"下载 {f_url} 的文件警告: 文件大小异常, 不到 0.25KB")

        chunk_size = 8192
        nowsize = 0
        succ = False
        lastime = time.time()
        useSpeed = 0

        with open(f_dir + ".tmp", "wb") as dwnf:
            for chk in res.iter_content(chunk_size=8192):
                nowtime = time.time()

                if nowtime != lastime:
                    useSpeed = chunk_size / (nowtime - lastime)
                    lastime = nowtime

                nowsize += len(chk)
                dwnf.write(chk)

                if nowsize % 81920 == 0:  # 每下载 10 个数据块更新一次进度
                    prgs = nowsize / filesize
                    _tmp = int(prgs * 20)
                    bar = Print.colormode_replace(
                        "§f" + " " * _tmp + "§b" + " " * (20 - _tmp) + "§r ", 7
                    )
                    Print.print_with_info(
                        f"{bar} {_pretty_kb(nowsize)}B / {_pretty_kb(filesize)}B ({_pretty_kb(useSpeed)}B/s)    ",
                        "§a 下载 §r",
                        end="\r",
                        need_log=False,
                    )
        succ = True

        if succ:
            shutil.move(f_dir + ".tmp", f_dir)
        else:
            os.remove(f_dir + ".tmp")


def download_unknown_file(url: str, save_dir: str):
    # 鉴于 Content-Length 不一定表示文件原始大小, 二进制文件与文本文件需要分开下载
    if _is_common_text_file(url) or _path_get_filename(url) in ("LICENSE",):
        resp = requests.get(url)
        resp.raise_for_status()
        with open(save_dir, "w", encoding="utf-8") as f:
            f.write(resp.text)
    else:
        download_file(url, save_dir)


def get_free_port(start=8080, end=65535):
    for port in range(start, end):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                return port
            except OSError:
                Print.print_war(f"端口 {port} 正被占用, 跳过")
    raise Exception(f"未找到空闲端口({start}~{end})")



import ctypes
import json
import os.path
import traceback
import threading
import enum
from typing import Iterable, Tuple, Optional, Union, Any, Callable, List, Dict
from threading import Thread
from dataclasses import dataclass
from ..color_print import Print

CInt = ctypes.c_longlong
CString = ctypes.c_char_p
CBytes = ctypes.POINTER(ctypes.c_char)


class byteCSlice(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.POINTER(ctypes.c_char)),
        ("len", ctypes.c_longlong),
        ("cap", ctypes.c_longlong),
    ]


def toCString(string: str):
    return ctypes.c_char_p(bytes(string, encoding="utf-8"))


def to_GoInt(i: int):
    return ctypes.c_longlong(i)


def toPyString(cstring: bytes):
    if cstring is None:
        return ""
    return cstring.decode(encoding="utf-8")


def toByteCSlice(bs: bytes):
    l = len(bs)
    kwargs = {
        "data": (ctypes.c_char * l)(*bs),
        "len": l,
        "cap": l,
    }
    return byteCSlice(**kwargs)


# define lib path and how to load it
import platform

sys_machine = platform.machine().lower()
sys_type = platform.uname().system
sys_fn = os.path.join(os.getcwd(), "tooldelta")
if sys_machine == "x86_64":
    sys_machine = "amd64"
elif sys_machine == "aarch64":
    sys_machine = "arm64"
if sys_type == "Windows":
    lib_path = f"neomega_windows_{sys_machine}.dll"
    lib_path = os.path.join(sys_fn, "neo_libs", lib_path)
    LIB = ctypes.cdll.LoadLibrary(lib_path)
elif "TERMUX_VERSION" in os.environ:
    lib_path = "neomega_android_arm64.so"
    lib_path = os.path.join(sys_fn, "neo_libs", lib_path)
    LIB = ctypes.CDLL(lib_path)
elif sys_type == "Linux":
    lib_path = f"neomega_linux_{sys_machine}.so"
    lib_path = os.path.join(sys_fn, "neo_libs", lib_path)
    LIB = ctypes.CDLL(lib_path)
else:
    lib_path = f"neomega_macos_{sys_machine}.dylib"
    lib_path = os.path.join(sys_fn, "neo_libs", lib_path)
    LIB = ctypes.CDLL(lib_path)

# define lib functions

# lib core functions: connect
LIB.ConnectOmega.argtypes = [CString]
LIB.ConnectOmega.restype = CString


def ConnectOmega(address):
    r = LIB.ConnectOmega(toCString(address))
    if r is not None:
        raise Exception(toPyString(r))


@dataclass
class AccountOptions:
    AuthServer: str = "https://api.fastbuilder.pro"
    UserToken: str = ""
    UserName: str = ""
    UserPassword: str = ""
    ServerCode: str = ""
    ServerPassword: str = ""


LIB.StartOmega.argtypes = [CString, CString]
LIB.StartOmega.restype = CString


def StartOmega(address, AccountOptions):
    r = LIB.StartOmega(
        toCString(address), toCString(json.dumps(AccountOptions.__dict__))
    )
    if r is not None:
        raise Exception(toPyString(r))


LIB.OmegaAvailable.restype = ctypes.c_uint8


def OmegaAvailable():
    if LIB.OmegaAvailable() != 1:
        raise Exception("omega Core disconnected")


# end lib core functions: connect


# lib core: event basic
class Event(ctypes.Structure):
    _fields_ = [("type", CString), ("retriever", CString)]


LIB.EventPoll.restype = Event


def EventPoll() -> tuple[str, str]:
    event = LIB.EventPoll()
    return toPyString(event.type), toPyString(event.retriever)


def OmitEvent():
    LIB.OmitEvent()


# end lib core: event

# event retrievers
LIB.ConsumeOmegaConnError.restype = CString
LIB.ConsumeCommandResponseCB.restype = CString


class MCPacketEvent(ctypes.Structure):
    _fields_ = [("packetDataAsJsonStr", CString), ("convertError", CString)]


LIB.ConsumeMCPacket.restype = MCPacketEvent

# Async Actions
# cmds

LIB.SendWebSocketCommandNeedResponse.argtypes = [CString, CString]


def SendWebSocketCommandNeedResponse(cmd: str, retrieverID: str):
    OmegaAvailable()
    LIB.SendWebSocketCommandNeedResponse(toCString(cmd), toCString(retrieverID))


LIB.SendPlayerCommandNeedResponse.argtypes = [CString, CString]


def SendPlayerCommandNeedResponse(cmd: str, retrieverID: str):
    OmegaAvailable()
    LIB.SendPlayerCommandNeedResponse(toCString(cmd), toCString(retrieverID))


# OneWay Actions
LIB.SendWOCommand.argtypes = [CString]


def SendSettingsCommand(cmd: str):
    OmegaAvailable()
    LIB.SendWOCommand(toCString(cmd))


LIB.SendWebSocketCommandOmitResponse.argtypes = [CString]


def SendWebSocketCommandOmitResponse(cmd: str):
    OmegaAvailable()
    LIB.SendWebSocketCommandOmitResponse(toCString(cmd))


LIB.SendPlayerCommandOmitResponse.argtypes = [CString]


def SendPlayerCommandOmitResponse(cmd: str):
    OmegaAvailable()
    LIB.SendPlayerCommandOmitResponse(toCString(cmd))


# Instance Actions
LIB.FreeMem.argtypes = [ctypes.c_void_p]
LIB.ListenAllPackets.argtypes = None
LIB.GetPacketNameIDMapping.restype = CString
LIB.JsonStrAsIsGamePacketBytes.argtypes = [CInt, CString]


class JsonStrAsIsGamePacketBytes_return(ctypes.Structure):
    _fields_ = [("pktBytes", CBytes), ("l", CInt), ("err", CString)]


LIB.JsonStrAsIsGamePacketBytes.restype = JsonStrAsIsGamePacketBytes_return


def JsonStrAsIsGamePacketBytes(packetID: int, jsonStr: str) -> bytes:
    r = LIB.JsonStrAsIsGamePacketBytes(to_GoInt(packetID), toCString(jsonStr))
    if toPyString(r.err) != "":
        raise ValueError(toPyString(r.err))
    bs = r.pktBytes[: r.l]
    LIB.FreeMem(r.pktBytes)
    return bs


LIB.SendGamePacket.argtypes = [CInt, CString]
LIB.SendGamePacket.restype = CString


def SendGamePacket(packetID: int, jsonStr: str) -> None:
    r = LIB.SendGamePacket(to_GoInt(packetID), toCString(jsonStr))
    if toPyString(r) != "":
        raise ValueError(toPyString(r))


LIB.GetClientMaintainedBotBasicInfo.restype = CString


# ClientMaintainedBotBasicInfo will not change
@dataclass
class ClientMaintainedBotBasicInfo:
    BotName: str = ""
    BotRuntimeID: int = 0
    BotUniqueID: int = 0
    BotIdentity: str = ""
    BotUUIDStr: str = ""


LIB.GetClientMaintainedExtendInfo.restype = CString


# any member of ClientMaintainedExtendInfo could be not found(which means related info not currently received from server)
@dataclass
class ClientMaintainedExtendInfo:
    CompressThreshold: Optional[int] = None
    WorldGameMode: Optional[int] = None
    WorldDifficulty: Optional[int] = None
    Time: Optional[int] = None
    DayTime: Optional[int] = None
    TimePercent: Optional[float] = None
    GameRules: Optional[Dict[str, Any]] = None


LIB.GetAllOnlinePlayers.restype = CString
LIB.ReleaseBindPlayer.argtypes = [CString]
LIB.PlayerName.argtypes = [CString]
LIB.PlayerName.restype = CString
LIB.PlayerEntityUniqueID.argtypes = [CString]
LIB.PlayerEntityUniqueID.restype = ctypes.c_int64
LIB.PlayerLoginTime.argtypes = [CString]
LIB.PlayerLoginTime.restype = ctypes.c_int64
LIB.PlayerPlatformChatID.argtypes = [CString]
LIB.PlayerPlatformChatID.restype = CString
LIB.PlayerBuildPlatform.argtypes = [CString]
LIB.PlayerBuildPlatform.restype = ctypes.c_int32
LIB.PlayerSkinID.argtypes = [CString]
LIB.PlayerSkinID.restype = CString
LIB.PlayerPropertiesFlag.argtypes = [CString]
LIB.PlayerPropertiesFlag.restype = ctypes.c_uint32
LIB.PlayerCommandPermissionLevel.argtypes = [CString]
LIB.PlayerCommandPermissionLevel.restype = ctypes.c_uint32
LIB.PlayerActionPermissions.argtypes = [CString]
LIB.PlayerActionPermissions.restype = ctypes.c_uint32
LIB.PlayerGetAbilityString.argtypes = [CString]
LIB.PlayerGetAbilityString.restype = CString
LIB.PlayerOPPermissionLevel.argtypes = [CString]
LIB.PlayerOPPermissionLevel.restype = ctypes.c_uint32
LIB.PlayerCustomStoredPermissions.argtypes = [CString]
LIB.PlayerCustomStoredPermissions.restype = ctypes.c_uint32
LIB.PlayerDeviceID.argtypes = [CString]
LIB.PlayerDeviceID.restype = CString
LIB.PlayerEntityRuntimeID.argtypes = [CString]
LIB.PlayerEntityRuntimeID.restype = ctypes.c_uint64
LIB.PlayerEntityMetadata.argtypes = [CString]
LIB.PlayerEntityMetadata.restype = CString
LIB.PlayerIsOP.argtypes = [CString]
LIB.PlayerIsOP.restype = ctypes.c_uint8
LIB.PlayerOnline.argtypes = [CString]
LIB.PlayerOnline.restype = ctypes.c_uint8

LIB.GetPlayerByUUID.argtypes = [CString]
LIB.GetPlayerByUUID.restype = CString

LIB.GetPlayerByName.argtypes = [CString]
LIB.GetPlayerByName.restype = CString

LIB.ConsumePlayerChange.restype = CString
LIB.InterceptPlayerJustNextInput.argtypes = [CString, CString]
LIB.ConsumeChat.restype = CString

LIB.PlayerChat.argtypes = [CString, CString]
LIB.PlayerTitle.argtypes = [CString, CString, CString]
LIB.PlayerActionBar.argtypes = [CString, CString]
LIB.SetPlayerAbility.argtypes = [CString, CString]

LIB.ListenCommandBlock.argtypes = [CString]
LIB.PlaceCommandBlock.argtypes = [CString]


def ConsumeChat() -> "Chat":
    chatData = json.loads(toPyString(LIB.ConsumeChat()))
    return Chat(**chatData)


class Counter:
    def __init__(self, prefix: str) -> None:
        self.current_i = 0
        self.prefix = prefix

    def __next__(self) -> int:
        self.current_i += 1
        return f"{self.prefix}_{self.current_i}"


@dataclass
class ActionPermissionMap:
    ActionPermissionAttackMobs: bool = False
    ActionPermissionAttackPlayers: bool = False
    ActionPermissionBuild: bool = False
    ActionPermissionDoorsAndSwitches: bool = False
    ActionPermissionMine: bool = False
    ActionPermissionOpenContainers: bool = False
    ActionPermissionOperator: bool = False
    ActionPermissionTeleport: bool = False
    ActionPermissionUnknown: bool = False


@dataclass
class AdventureFlagsMap:
    AdventureFlagAllowFlight: bool = False
    AdventureFlagAutoJump: bool = False
    AdventureFlagFlying: bool = False
    AdventureFlagMuted: bool = False
    AdventureFlagNoClip: bool = False
    AdventureFlagWorldBuilder: bool = False
    AdventureFlagWorldImmutable: bool = False
    AdventureSettingsFlagsNoMvP: bool = False
    AdventureSettingsFlagsNoPvM: bool = False
    AdventureSettingsFlagsShowNameTags: bool = False


@dataclass
class Chat:
    Name: str = ""
    Msg: List[str] = ""
    Type: int = 1
    RawMsg: str = ""
    RawName: str = ""
    RawParameters: Optional[Any] = None
    Aux: Optional[Any] = None


@dataclass
class CommandOrigin:
    Origin: int = 0
    UUID: str = ""
    RequestID: str = ""
    PlayerUniqueID: int = 0


@dataclass
class OutputMessage:
    Success: bool = False
    Message: str = ""
    Parameters: Optional[List[Any]] = None


@dataclass
class CommandOutput:
    CommandOrigin: Optional[CommandOrigin] = None
    OutputType: int = 0
    SuccessCount: int = 0
    OutputMessages: Optional[List[OutputMessage]] = None
    DataSet: Optional[Any] = None

    @property
    def as_dict(self):
        return _unpack_command_output(self)


def unpackCommandOutput(jsonStr: Optional[str]) -> Optional[CommandOutput]:
    if jsonStr is None:
        return None
    commandOutputData = json.loads(jsonStr)
    commandOrigin = CommandOrigin(**commandOutputData["CommandOrigin"])
    dataset = None
    if (
        "DataSet" in commandOutputData.keys()
        and commandOutputData["DataSet"] is not None
        and commandOutputData["DataSet"] != ""
    ):
        dataset = json.loads(commandOutputData["DataSet"])
    outputMessages = []
    if (
        "OutputMessages" in commandOutputData.keys()
        and commandOutputData["OutputMessages"] is not None
        and commandOutputData["OutputMessages"] != ""
    ):
        outputMessagesData = commandOutputData["OutputMessages"]
        for outputMessageData in outputMessagesData:
            outputMessage = OutputMessage(
                Success=outputMessageData["Success"],
                Message=outputMessageData["Message"],
            )
            parameters = []
            if "Parameters" in outputMessageData.keys():
                parametersData = outputMessageData["Parameters"]
                for parameterData in parametersData:
                    try:
                        parameters.append(json.loads(parameterData))
                    except:
                        parameters.append(parameterData)
            outputMessage.Parameters = parameters
            outputMessages.append(outputMessage)
    return CommandOutput(
        CommandOrigin=commandOrigin,
        OutputType=commandOutputData["OutputType"],
        SuccessCount=commandOutputData["SuccessCount"],
        OutputMessages=outputMessages,
        DataSet=dataset,
    )


@dataclass
class CommandBlockPlaceOption:
    X: int = 0
    Y: int = 0
    Z: int = 0
    BlockName: str = ""
    BockState: str = ""
    NeedRedStone: bool = False
    Conditional: bool = False
    Command: str = ""
    Name: str = ""
    TickDelay: int = 0
    ShouldTrackOutput: bool = False
    ExecuteOnFirstTick: bool = False


@dataclass
class QueriedPlayerPos:
    dimension: int = 0
    x: float = 0
    y: float = 0.0
    z: float = 0.0
    yRot: float = 0.0


class PlayerKit:
    def __init__(self, uuid: str, parent: "ThreadOmega") -> None:
        self.parent = parent
        self._uuid = uuid
        self._c_uuid = toCString(self._uuid)

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def name(self) -> str:
        OmegaAvailable()
        return toPyString(LIB.PlayerName(self._c_uuid))

    @property
    def entity_unique_id(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerEntityUniqueID(self._c_uuid))

    @property
    def op(self) -> bool:
        OmegaAvailable()
        return LIB.PlayerIsOP(self._c_uuid) == 1

    @property
    def online(self) -> bool:
        OmegaAvailable()
        return LIB.PlayerOnline(self._c_uuid) == 1

    @property
    def login_time(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerLoginTime(self._c_uuid))

    @property
    def platform_chat_id(self) -> str:
        OmegaAvailable()
        return toPyString(LIB.PlayerPlatformChatID(self._c_uuid))

    @property
    def build_platform(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerBuildPlatform(self._c_uuid))

    @property
    def skin_id(self) -> str:
        OmegaAvailable()
        return toPyString(LIB.PlayerSkinID(self._c_uuid))

    @property
    def device_id(self) -> str:
        OmegaAvailable()
        return toPyString(LIB.PlayerDeviceID(self._c_uuid))

    @property
    def properties_flag(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerPropertiesFlag(self._c_uuid))

    @property
    def command_permission_level(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerCommandPermissionLevel(self._c_uuid))

    @property
    def action_permissions(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerActionPermissions(self._c_uuid))

    @property
    def op_permissions_level(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerOPPermissionLevel(self._c_uuid))

    @property
    def custom_permissions(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerCustomStoredPermissions(self._c_uuid))

    @property
    def ability_map(self) -> Tuple[ActionPermissionMap, AdventureFlagsMap]:
        OmegaAvailable()
        maps = json.loads(toPyString(LIB.PlayerGetAbilityString(self._c_uuid)))
        ActionPermissionMapData = maps["ActionPermissionMap"]
        AdventureFlagsMapData = maps["AdventureFlagsMap"]
        return ActionPermissionMap(**ActionPermissionMapData), AdventureFlagsMap(
            **AdventureFlagsMapData
        )

    @property
    def entity_runtime_id(self) -> int:
        OmegaAvailable()
        return int(LIB.PlayerEntityRuntimeID(self._c_uuid))

    @property
    def entity_metadata(self) -> any:
        OmegaAvailable()
        return json.loads(toPyString(LIB.PlayerEntityMetadata(self._c_uuid)))

    def say(self, msg: str):
        OmegaAvailable()
        LIB.PlayerChat(self._c_uuid, toCString(msg))

    def ask(self, hint: str) -> str:
        OmegaAvailable()
        self.say(hint)
        return self.parent.intercept_player_just_next_input(self._c_uuid).RawMsg.strip()

    def title(self, title: str = "", subtitle: str = ""):
        # subtitle 只有在 title 给出时才能实际生效
        LIB.PlayerTitle(self._c_uuid, toCString(title), toCString(subtitle))

    def action_bar(self, msg: str):
        LIB.PlayerActionBar(self._c_uuid, toCString(msg))

    def query(self, conditions: Union[None, str, List[str]] = None) -> CommandOutput:
        query_str = f'querytarget @a[name="{self.name}"'
        if conditions is None:
            query_str += "]"
        elif isinstance(conditions, str):
            query_str += "," + conditions + "]"
        else:
            query_str += "," + ",".join(conditions) + "]"
        ret = self.parent.send_websocket_command_need_response(query_str)
        return ret

    def check_conditions(self, conditions: Union[None, str, List[str]] = None) -> bool:
        return self.query(conditions).SuccessCount > 0

    def get_pos(self) -> QueriedPlayerPos:
        pos = self.query().OutputMessages[0].Parameters[0][0]
        return QueriedPlayerPos(
            dimension=pos["dimension"],
            yRot=pos["yRot"],
            x=pos["position"]["x"],
            y=pos["position"]["y"],
            z=pos["position"]["z"],
        )

    def set_ability_map(
        self, action_permission: ActionPermissionMap, adventure_flag: AdventureFlagsMap
    ):
        jsonFlags = json.dumps(
            {
                "AdventureFlagsMap": adventure_flag.__dict__,
                "ActionPermissionMap": action_permission.__dict__,
            }
        )
        LIB.SetPlayerAbility(self._c_uuid, toCString(jsonFlags))

    def __repr__(self) -> str:
        return f"uuid={self.uuid},name={self.name},entity_unique_id={self.entity_unique_id},op={self.op},online={self.online}"

    def __del__(self):
        LIB.ReleaseBindPlayer(toCString(self._uuid))


class ConnectType(enum.Enum):
    Remote = "Remote"  # 连接到一个 neOmega Access Point
    Local = "Local"  # 在内部启动一个单独的 neOmega Core


class ThreadOmega:
    def __init__(
        self,
        connect_type: ConnectType,
        address: str = "tcp://localhost:24016",
        accountOption: AccountOptions = None,
    ) -> None:
        self._thread_counter = Counter("thread")
        self._running_threads: Dict[str, Thread] = {}
        if connect_type == ConnectType.Local:
            StartOmega(address, accountOption)
            Print.print_inf(f"Omega 接入点已启动, 在 {address} 开放接口")
        elif connect_type == ConnectType.Remote:
            ConnectOmega(address)

        # disconnect event
        self._omega_disconnected_lock = threading.Event()
        self._omega_disconnected_lock.clear()  # lock
        self._omega_disconnected_reason = ""

        # cmd events
        self._cmd_callback_retriever_counter = Counter("cmd_callback")
        self._omega_cmd_callback_events: Dict[str, Callable] = {}

        # packet listeners
        self._packet_listeners: Dict[str, List[Callable[[str, any], None]]] = {}

        # setup actions
        LIB.ListenAllPackets()  # make LIB listen to all packets and new packets will have eventType="MCPacket"
        mapping = json.loads(toPyString(LIB.GetPacketNameIDMapping()))
        self._packet_name_to_id_mapping: dict[str, int] = mapping
        self._packet_id_to_name_mapping = {}
        for packet_name, packet_id in self._packet_name_to_id_mapping.items():
            self._packet_id_to_name_mapping[packet_id] = packet_name
            self._packet_listeners[packet_name] = []

        LIB.ListenPlayerChange()
        self._player_change_listeners: List[Callable[[PlayerKit, str], None]] = []

        # get bot basic info (this info will not change so we need to get it only once)
        self._bot_basic_info = ClientMaintainedBotBasicInfo(
            **json.loads(toPyString(LIB.GetClientMaintainedBotBasicInfo()))
        )

        # player hooks
        self._bind_players: Dict[str, PlayerKit] = {}

        # player chat intercept callbacks
        self._player_chat_intercept_callback_retriever_counter = Counter(
            "player_chat_intercept"
        )
        self._player_chat_intercept_callback_events: Dict[str, Callable] = {}

        # chat
        LIB.ListenChat()
        self._player_chat_listeners: List[Callable[[Chat, PlayerKit], None]] = []
        self._specific_chat_listeners: Dict[
            str, List[Callable[[Chat, PlayerKit], None]]
        ] = {}

        # named command block msg
        self._name_command_block_msg_listeners: Dict[
            str, List[Callable[[Chat], None]]
        ] = {}

        # start routine
        self.start_new(self._react)

    def start_new(self, func: Callable, args: Iterable[Any] = ()):
        thread_i = next(self._thread_counter)

        def clean_up(*args):
            try:
                func(*args)
            except Exception:
                print(traceback.print_exc(limit=None, file=None, chain=True))
            finally:
                del self._running_threads[thread_i]

        t = Thread(target=clean_up, args=args)
        self._running_threads[thread_i] = t
        t.start()

    def _react(self):
        while True:
            eventType, retriever = EventPoll()
            if eventType == "OmegaConnErr":
                self._omega_disconnected_reason = toPyString(
                    LIB.ConsumeOmegaConnError()
                )
                self._omega_disconnected_lock.set()
            elif eventType == "CommandResponseCB":
                cmdResp = unpackCommandOutput(
                    toPyString(LIB.ConsumeCommandResponseCB())
                )
                self._omega_cmd_callback_events[retriever](cmdResp)
            elif eventType == "MCPacket":
                packetTypeName = retriever
                if packetTypeName == "":
                    print("'', ignored")
                listeners = self._packet_listeners.get(packetTypeName, [])
                if len(listeners) == 0:
                    LIB.OmitEvent()
                else:
                    ret = LIB.ConsumeMCPacket()
                    if toPyString(ret.convertError) != "":
                        raise ValueError(toPyString(ret.convertError))
                    jsonPkt = json.loads(toPyString(ret.packetDataAsJsonStr))
                    for listener in listeners:
                        self.start_new(listener, (packetTypeName, jsonPkt))
            elif eventType == "PlayerChange":
                playerUUID = retriever
                if len(self._player_change_listeners) == 0:
                    LIB.OmitEvent()
                else:
                    action = toPyString(LIB.ConsumePlayerChange())
                    for callback in self._player_change_listeners:
                        self.start_new(
                            callback, (self._get_bind_player(playerUUID), action)
                        )

            elif eventType == "PlayerInterceptInput":
                chat = ConsumeChat()
                self._player_chat_intercept_callback_events[retriever](chat)

            elif eventType == "Chat":
                chat = ConsumeChat()
                if (
                    not self._player_chat_listeners
                    and not self._specific_chat_listeners.get(chat.Name)
                    and not self._specific_chat_listeners.get(chat.RawName)
                ):
                    LIB.OmitEvent()
                else:
                    player = self.get_player_by_name(chat.Name)
                    if not player:
                        name = chat.Name
                        if name in self._specific_chat_listeners:
                            for callback in self._specific_chat_listeners[name]:
                                self.start_new(callback, (chat,))
                        if chat.Name != chat.RawName:
                            name = chat.RawName
                            if name in self._specific_chat_listeners:
                                for callback in self._specific_chat_listeners[name]:
                                    self.start_new(callback, (chat,))
                    else:
                        for callback in self._player_chat_listeners:
                            self.start_new(callback, (chat, player))

            elif eventType == "NamedCommandBlockMsg":
                blockName = retriever
                listeners = self._name_command_block_msg_listeners[blockName]
                if len(listeners) == 0:
                    LIB.OmitEvent()
                else:
                    chat = ConsumeChat()
                    for l in listeners:
                        self.start_new(l, (chat,))

    def wait_disconnect(self) -> str:
        """return: disconnect reason"""
        self._omega_disconnected_lock.wait()
        return self._omega_disconnected_reason

    @staticmethod
    def _create_lock_and_result_setter():
        lock = threading.Lock()
        lock.acquire()
        ret = [None]

        def result_setter(result):
            ret[0] = result
            lock.release()

        def result_getter(timeout: int = -1) -> Any:
            lock.acquire(timeout=timeout)
            return ret[0]

        return result_setter, result_getter

    def send_websocket_command_need_response(
        self, cmd: str, timeout: int = -1
    ) -> Optional[CommandOutput]:
        setter, getter = self._create_lock_and_result_setter()
        retriever_id = next(self._cmd_callback_retriever_counter)
        self._omega_cmd_callback_events[retriever_id] = setter
        SendWebSocketCommandNeedResponse(cmd, retriever_id)
        res = getter(timeout=timeout)
        del self._omega_cmd_callback_events[retriever_id]
        return res

    def send_player_command_need_response(
        self, cmd: str, timeout: int = -1
    ) -> Optional[CommandOutput]:
        setter, getter = self._create_lock_and_result_setter()
        retriever_id = next(self._cmd_callback_retriever_counter)
        self._omega_cmd_callback_events[retriever_id] = setter
        SendPlayerCommandNeedResponse(cmd, retriever_id)
        res = getter(timeout=timeout)
        del self._omega_cmd_callback_events[retriever_id]
        return res

    @staticmethod
    def send_settings_command(cmd: str):
        SendSettingsCommand(cmd)

    @staticmethod
    def send_websocket_command_omit_response(cmd: str):
        SendWebSocketCommandOmitResponse(cmd)

    @staticmethod
    def send_player_command_omit_response(cmd: str):
        SendPlayerCommandOmitResponse(cmd)

    def get_packet_name_to_id_mapping(
        self, requires: Optional[Union[List[str], str]] = None
    ) -> Union[Dict[str, int], int]:
        if requires is None:
            return dict(self._packet_name_to_id_mapping.items())
        if isinstance(requires, list):
            return {k: self._packet_name_to_id_mapping[k] for k in requires}
        return self._packet_name_to_id_mapping[requires]

    def get_packet_id_to_name_mapping(
        self, requires: Optional[Union[List[int], int]] = None
    ) -> Union[Dict[int, str], str]:
        if requires is None:
            return dict(self._packet_id_to_name_mapping.items())
        if isinstance(requires, list):
            return {k: self._packet_id_to_name_mapping[k] for k in requires}
        return self._packet_id_to_name_mapping[requires]

    def listen_packets(
        self, targets: Union[str, List[str]], callback: Callable[[str, Any], None]
    ):
        if isinstance(targets, str):
            targets = [targets]
        if isinstance(targets, int):
            targets = [f"{targets}"]
        translate_targets = {}
        for t in targets:
            hit = False
            if t == "all":
                translate_targets = {
                    k: True for k in self.get_packet_name_to_id_mapping().keys()
                }
                hit = True
            elif t.startswith("!"):
                rt = t[1:]
                if rt in self._packet_name_to_id_mapping.keys():
                    translate_targets[rt] = False
                    hit = True
                try:
                    packetID = int(rt)
                    packetType = self._packet_id_to_name_mapping(packetID)
                    translate_targets[packetType] = False
                    hit = True
                except:
                    pass
            else:
                if t in self._packet_name_to_id_mapping.keys():
                    translate_targets[t] = True
                    hit = True
                try:
                    packetID = int(t)
                    packetType = self.get_packet_id_to_name_mapping(packetID)
                    translate_targets[packetType] = True
                    hit = True
                except:
                    pass
            if not hit:
                raise ValueError(f"{t}")
        for k, v in translate_targets.items():
            if v:
                self._packet_listeners[k].append(callback)

    def construct_game_packet_bytes_in_json_as_is(
        self, packet_type: Union[int, str], content: Any
    ) -> tuple[int, bytes]:
        if isinstance(packet_type, str):
            packet_type = self.get_packet_name_to_id_mapping(packet_type)
        return packet_type, JsonStrAsIsGamePacketBytes(packet_type, json.dumps(content))

    def send_game_packet_in_json_as_is(
        self, packet_type: Union[int, str], content: Any
    ):
        if isinstance(packet_type, str):
            packet_type = self.get_packet_name_to_id_mapping(packet_type)
        OmegaAvailable()
        SendGamePacket(packet_type, json.dumps(content))

    def get_bot_basic_info(self) -> ClientMaintainedBotBasicInfo:
        return self._bot_basic_info

    def get_bot_name(self) -> str:
        return self._bot_basic_info.BotName

    def get_bot_runtime_id(self) -> int:
        return self._bot_basic_info.BotRuntimeID

    def get_bot_unique_id(self) -> int:
        return self._bot_basic_info.BotUniqueID

    def get_bot_identity(self) -> str:
        return self._bot_basic_info.BotIdentity

    def get_bot_uuid_str(self) -> str:
        return self._bot_basic_info.BotUUIDStr

    @staticmethod
    def get_extend_info() -> ClientMaintainedExtendInfo:
        OmegaAvailable()
        return ClientMaintainedExtendInfo(
            **json.loads(toPyString(LIB.GetClientMaintainedExtendInfo()))
        )

    def _get_bind_player(self, uuidStr: str) -> Optional[PlayerKit]:
        if uuidStr is None or uuidStr == "":
            return None
        if uuidStr in self._bind_players:
            return self._bind_players[uuidStr]
        bind_player = PlayerKit(uuidStr, self)
        self._bind_players[uuidStr] = bind_player
        return bind_player

    def get_all_online_players(self):
        OmegaAvailable()
        playerUUIDS = json.loads(toPyString(LIB.GetAllOnlinePlayers()))
        return [self._get_bind_player(uuidStr) for uuidStr in playerUUIDS]

    def get_player_by_name(self, name: str) -> Optional[PlayerKit]:
        OmegaAvailable()
        playerUUID = toPyString(LIB.GetPlayerByName(toCString(name)))
        return self._get_bind_player(playerUUID)

    def get_player_by_uuid(self, uuidStr: str) -> Optional[PlayerKit]:
        OmegaAvailable()
        playerUUID = toPyString(LIB.GetPlayerByUUID(toCString(uuidStr)))
        return self._get_bind_player(playerUUID)

    def listen_player_change(self, callback: Callable[[PlayerKit, str], None]):
        for player in self.get_all_online_players():
            callback(player, "exist")
        self._player_change_listeners.append(callback)

    def intercept_player_just_next_input(
        self, player_c_uuid: CString, timeout: int = -1
    ) -> Chat:
        setter, getter = self._create_lock_and_result_setter()
        retrieverID = next(self._player_chat_intercept_callback_retriever_counter)
        self._player_chat_intercept_callback_events[retrieverID] = setter
        LIB.InterceptPlayerJustNextInput(player_c_uuid, toCString(retrieverID))
        res = getter(timeout=timeout)
        del self._player_chat_intercept_callback_events[retrieverID]
        return res

    def listen_player_chat(self, callback: Callable[[Chat, PlayerKit], None]):
        self._player_chat_listeners.append(callback)

    def listen_specific_chat(
        self, specific_name: str, callback: Callable[[Chat], None]
    ):
        if specific_name not in self._specific_chat_listeners:
            self._specific_chat_listeners[specific_name] = []
        self._specific_chat_listeners[specific_name].append(callback)

    def listen_named_command_block(
        self, command_block_name: str, callback: Callable[[Chat], None]
    ):
        if command_block_name not in self._name_command_block_msg_listeners:
            self._name_command_block_msg_listeners[command_block_name] = []
        LIB.ListenCommandBlock(toCString(command_block_name))
        self._name_command_block_msg_listeners[command_block_name].append(callback)

    @staticmethod
    def place_command_block(place_option: CommandBlockPlaceOption):
        LIB.PlaceCommandBlock(toCString(json.dumps(place_option.__dict__)))

    def __del__(self):
        for t in self._running_threads.values():
            t.join()


def _unpack_command_output(c: CommandOutput):
    return {
        "CommandOrigin": {
            "Orogin": c.CommandOrigin.Origin,
            "UUID": c.CommandOrigin.UUID,
            "RequestID": c.CommandOrigin.RequestID,
            "PlayerUniqueID": c.CommandOrigin.PlayerUniqueID,
        },
        "OutputMessages": [_unpack_output_msgs(i) for i in c.OutputMessages],
        "OutputType": c.OutputType,
        "SuccessCOunt": c.SuccessCount,
        "DataSet": c.DataSet,
    }


def _unpack_output_msgs(c: OutputMessage):
    return {"Success": c.Success, "Message": c.Message, "Parameters": c.Parameters}



import time
from tooldelta import builtins
from tooldelta.Frame import PRG_NAME, Frame
from tooldelta.Frame import GameCtrl
from tooldelta.basic_mods import os, traceback
from tooldelta.color_print import Print
from tooldelta.plugin_load.PluginGroup import PluginGroup
from tooldelta.plugin_load.injected_plugin import movent

frame = Frame()
plugins = PluginGroup(frame, PRG_NAME)
game_control = GameCtrl(frame)


def start_tool_delta(exit_directly=True):
    global frame
    # 初始化系统
    try:
        frame.auto_update()
        frame.welcome()
        frame.basic_operation()
        frame.set_game_control(game_control)
        frame.set_plugin_group(plugins)
        movent.set_frame(frame)
        frame.read_cfg()
        game_control.init_funcs()
        plugins.read_all_plugins()
        frame.plugin_load_finished(plugins)
        plugins.execute_def(frame.on_plugin_err)
        builtins.tmpjson_save_thread()
        frame.launcher.listen_launched(game_control.Inject)
        game_control.set_listen_packets()
        raise frame.launcher.launch()
    except (KeyboardInterrupt, SystemExit):
        pass
    except:
        Print.print_err("ToolDelta 运行过程中出现问题: " + traceback.format_exc())


def safe_jump(exit_directly=True):
    frame.safe_close()
    if exit_directly:
        for _ in range(3, 0, -1):
            Print.print_war(f"{_}秒后强制退出...", end="\r")
            time.sleep(1)
        Print.print_war(f"0秒后强制退出...", end="\r")
        Print.print_suc("ToolDelta 已退出.")
        os._exit(0)
    Print.print_suc("ToolDelta 已退出.")



import os
import time


class ToolDeltaLogger:
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    FATAL = "FATAL"
    OTHER_TYPE = "???"

    def __init__(self, log_path, name_fmt="%Y-%m-%d"):
        self.path = log_path
        self.name_fmt = name_fmt
        self.now_day = time.strftime("%Y-%m-%d")
        self.logging_fmt = "[%H-%M-%S]"
        self.lastLogTime = time.time()
        self.open_wrapper_io(log_path)
        self.enable_logger = False
        self.writable = True

    def switch_logger(self, isopen: bool):
        self.enable_logger = isopen

    def open_wrapper_io(self, log_path: str):
        self._wrapper = open(
            log_path + os.sep + time.strftime(self.name_fmt) + ".log",
            "a",
            encoding="utf-8",
            buffering=4096,
        )

    def log_in(self, msg, level=INFO):
        # 写入日志信息. level给定了其等级.
        if not self.writable or not self.enable_logger:
            return
        if not isinstance(msg, str):
            raise TypeError("only allows string")
        if "\n" in msg:
            msg = msg.replace("\n", "\n    ")
        if len(msg) > 200:
            msg = msg[:200] + "..."
        self._check_is_another_day()
        self._wrapper.write(
            time.strftime(self.logging_fmt)
            + f" [{level}] "
            + (msg if msg.endswith("\n") else msg + "\n")
        )
        if time.time() - self.lastLogTime > 15:
            self._save_log()
            self.lastLogTime = time.time()

    def _save_log(self):
        self._wrapper.flush()

    def _check_is_another_day(self):
        # 判断记录日志的时候是否已经是第二天, 是的话就变更文件名.
        if time.strftime("%Y-%m-%d") != self.now_day:
            self.exit()
            self.open_wrapper_io(self.path)

    def exit(self):
        if self.writable:
            self.writable = False
            self._save_log()
            self._wrapper.close()


def new_logger(log_path: str):
    os.makedirs(log_path, exist_ok=True)
    return ToolDeltaLogger(log_path)


publicLogger = new_logger("日志文件")



import os
import json
import threading
import ctypes
import traceback


def get_dotcs_env(__F, print_ins):
    sendcmd = lambda cmd, waitForResp=False, timeout=30: (
        __F.link_game_ctrl.sendcmd(cmd, waitForResp, timeout).as_dict
        if waitForResp
        else __F.link_game_ctrl.sendcmd(cmd)
    )
    sendwocmd = __F.link_game_ctrl.sendwocmd
    sendwscmd = lambda cmd, waitForResp=False, timeout=30: (
        __F.link_game_ctrl.sendwscmd(cmd, waitForResp, timeout).as_dict
        if waitForResp
        else __F.link_game_ctrl.sendwscmd(cmd)
    )
    sendfbcmd = __F.link_game_ctrl.sendfbcmd
    allplayers = __F.link_game_ctrl.allplayers
    robotname = __F.link_game_ctrl.bot_name
    XUID2playerName = __F.link_game_ctrl.players_uuid
    threadList = __F.old_dotcs_threadinglist
    admin = adminhigh = [robotname]
    tellrawText = lambda target, dispname=None, text="": __F.link_game_ctrl.say_to(
        target, dispname + " " + text if dispname else text
    )
    exiting = False
    server = hash(__F.serverNumber)

    def color(
        text: str,
        output: bool = True,
        end: str = "\n",
        replace: bool = False,
        replaceByNext: bool = False,
        info=" 信息 ",
    ):
        if not output:
            return print_ins.fmt_info(text, info)
        if not replace:
            print_ins.print_with_info(text, info, end=end)
        else:
            print_ins.print_with_info(text, info, end="\r")

    class createThread(threading.Thread):
        def __init__(self, name, data=None, func="", output=True):
            if data is None:
                data = {}
            threading.Thread.__init__(self)
            self.name = name
            self.data = data
            self.func = func
            self.stopping = False
            self.daemon = True
            self.output = output
            threadList.append(self)
            self.start()

        def run(self):
            try:
                if self.func.__class__.__name__ != "str":
                    self.func(self)
                else:
                    exec("%s(self)" % self.func)
            except Exception as err:
                traceback.print_exc()
            except SystemExit as err:
                pass
            finally:
                threadList.remove(self)

        def get_id(self):
            if hasattr(self, "_thread_id"):
                return self._thread_id
            for thread_id, thread in enumerate(threading.enumerate()):
                if thread is self:
                    return thread_id

        def stop(self):
            self.stopping = True
            thread_id = self.get_id()
            res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
                thread_id, ctypes.py_object(SystemExit)
            )
            if res > 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)

    def getTarget(sth: str, timeout: bool | int = 1) -> list:
        if not sth.startswith("@"):
            raise Exception("Minecraft Target Selector is not correct.")
        result = sendcmd("/tell @s get%s" % sth, True, timeout)["OutputMessages"][0][
            "Parameters"
        ][1][3:]
        if ", " not in result:
            if not result:
                return []
            return [result]
        return result.split(", ")

    def getScore(scoreboardNameToGet: str, targetNameToGet: str) -> int:
        resultList = sendcmd("/scoreboard players list %s" % targetNameToGet, True)[
            "OutputMessages"
        ]
        result = {}
        result2 = {}
        for i in resultList:
            Message = i["Message"]
            if Message == r"commands.scoreboard.players.list.player.empty":
                continue
            if Message == r"§a%commands.scoreboard.players.list.player.count":
                targetName = i["Parameters"][1][1:]
            elif Message == "commands.scoreboard.players.list.player.entry":
                if targetName == "commands.scoreboard.players.offlinePlayerName":
                    continue
            scoreboardName = i["Parameters"][2]
            targetScore = int(i["Parameters"][0])
            if targetName not in result:
                result[targetName] = {}
            result[targetName][scoreboardName] = targetScore
            if scoreboardName not in result2:
                result2[scoreboardName] = {}
            result2[scoreboardName][targetName] = targetScore
        if not (result or result2):
            raise Exception("Failed to get the score.")
        try:
            if targetNameToGet == "*" or targetNameToGet.startswith("@"):
                if scoreboardNameToGet == "*":
                    return [result, result2]
                return result2[scoreboardNameToGet]
            if scoreboardNameToGet == "*":
                return result[targetNameToGet]
            return result[targetNameToGet][scoreboardNameToGet]
        except KeyError as err:
            raise Exception("Failed to get score: %s" % str(err))

    def getPos(targetNameToGet: str, timeout: float | int = 1) -> dict:
        if (
            (targetNameToGet not in allplayers)
            and (targetNameToGet != robotname)
            and (not targetNameToGet.startswith("@a"))
        ):
            raise Exception("Player not found.")
        result = sendcmd("/querytarget %s" % targetNameToGet, True, timeout)[
            "OutputMessages"
        ][0]
        if result["Success"] is False:
            raise Exception("Failed to get the position.")
        resultList = json.loads(result["Parameters"][0])
        result = {}
        for i in resultList:
            targetName = XUID2playerName[i["uniqueId"][-8:]]
            x = i["position"]["x"]
            y = i["position"]["y"] - 1.6200103759765
            z = i["position"]["z"]
            position = {
                "x": float("%.2f" % x),
                "y": float("%.2f" % y),
                "z": float("%.2f" % z),
            }
            dimension = i["dimension"]
            yRot = i["yRot"]
            result[targetName] = {
                "dimension": dimension,
                "position": position,
                "yRot": yRot,
            }
        if targetNameToGet == "@a":
            return result
        if len(result) != 1:
            raise Exception("Failed to get the position.")
        if targetNameToGet.startswith("@a"):
            return list(result.values())[0]
        return result[targetNameToGet]

    def getItem(targetName: str, itemName: str, itemSpecialID: int = -1) -> int:
        if (
            (targetName not in allplayers)
            and (targetName != robotname)
            and (not targetName.startswith("@a"))
        ):
            raise Exception("Player not found.")
        result = sendcmd(
            "/clear %s %s %d 0" % (targetName, itemName, itemSpecialID), True
        )
        if result["OutputMessages"][0]["Message"] == "commands.generic.syntax":
            raise Exception("Item name error.")
        if result["OutputMessages"][0]["Message"] == "commands.clear.failure.no.items":
            return 0
        return int(result["OutputMessages"][0]["Parameters"][1])

    def getStatus(statusName: str):
        if not os.path.isfile("data/dotcs_status_%s.txt" % statusName):
            return None
        with open(
            "data/dotcs_status_%s.txt" % statusName, "r", encoding="utf-8"
        ) as file:
            status = file.read()
        return status

    def setStatus(statusName: str, status):
        with open(  # skipcq: PTC-W6004
            "data/dotcs_status_%s.txt" % statusName, "w", encoding="utf-8"
        ) as file:
            file.write(str(status))

    def getPlayerData(dataName: str, playerName: str, writeNew):
        try:
            with open(
                f"data/players/{playerName}.json", "r", encoding="utf-8"
            ) as f:  # skipcq: PTC-W6004
                j = json.load(f)[dataName]
            return j
        except KeyError:
            with open(f"data/players/{playerName}.json", "r", encoding="utf-8") as f:
                j = json.load(f)
            j[dataName] = writeNew
            with open(f"data/players/{playerName}.json", "w", encoding="utf-8") as f:
                json.dump(j, f)
        except FileNotFoundError:
            with open(f"data/players/{playerName}.json", "w", encoding="utf-8") as f:
                json.dump({dataName: writeNew}, f)
        return None

    def setPlayerData(dataName: str, playerName: str, dataValue, writeNew: str = ""):
        if os.path.isfile(f"data/players/{playerName}.json"):
            with open(f"data/players/{playerName}.json", "r", encoding="utf-8") as f:
                j = json.load(f)
            j[dataName] = dataValue
            with open(f"data/players/{playerName}.json", "w", encoding="utf-8") as f:
                json.dump(j, f)
        else:
            with open(f"data/players/{playerName}.json", "w", encoding="utf-8") as f:
                json.dump({dataName: dataValue}, f)

    return locals()



import platform
import os
import shlex
import subprocess
import time
import json
import requests
import ujson
import random
from typing import Callable, Optional

from .color_print import Print
from .urlmethod import download_file, get_free_port
from .builtins import Builtins
from .packets import Packet_CommandOutput, PacketIDS
from .sys_args import sys_args_to_dict

# from .neo_libs import neo_conn
from . import fbconn
import threading

import tooldelta


class SysStatus:
    LOADING = 100
    LAUNCHING = 101
    RUNNING = 102
    NORMAL_EXIT = 103
    FB_LAUNCH_EXC = 104
    FB_CRASHED = 105
    NEED_RESTART = 106
    launch_type = "None"


class StandardFrame:
    # 提供了标准的启动器框架, 作为 ToolDelta 和游戏交互的接口
    launch_type = "Original"

    def __init__(self, serverNumber, password, fbToken, auth_server_url):
        self.serverNumber = serverNumber
        self.serverPassword = password
        self.fbToken = fbToken
        self.auth_server = auth_server_url
        self.status = SysStatus.LOADING
        self.system_type = platform.uname().system
        self.inject_events = []
        self.packet_handler = None
        self.need_listen_packets = {9, 63, 79}
        self._launcher_listener = None
        self.exit_event = threading.Event()

    def add_listen_packets(self, *pcks: int):
        for i in pcks:
            self.need_listen_packets.add(i)

    @staticmethod
    def launch():
        raise Exception("Cannot launch this launcher")

    def listen_launched(self, cb):
        self._launcher_listener = cb

    def close_fb(self):
        raise NotImplementedError()

    @staticmethod
    def get_players_and_uuids():
        return None

    @staticmethod
    def get_bot_name():
        return None

    get_all_players = None
    sendcmd: Callable[
        [str, Optional[bool], Optional[int | float]], None | Packet_CommandOutput
    ]
    sendwscmd: Callable[
        [str, Optional[bool], Optional[int | float]], Packet_CommandOutput
    ]
    sendwocmd: Callable[[str], None]
    sendfbcmd: Callable[[str], None | AttributeError]
    sendPacket: Callable[[int, str], None]
    sendPacketJson: Callable[[int, str], None]
    is_op: Callable[[str], None | bool] | None


class FrameFBConn(StandardFrame):
    # 使用原生 FastBuilder External 连接
    cmds_reqs = []
    cmds_resp = {}

    def __init__(self, serverNumber, password, fbToken, auth_server):
        super().__init__(serverNumber, password, fbToken, auth_server)
        self.injected = False
        self.downloadMissingFiles()
        self.init_all_functions()

    def launch(self):
        try:
            free_port = get_free_port(10000)
            self.runFB(port=free_port)
            self.run_conn(port=free_port)
            Builtins.createThread(self.output_fb_msgs_thread)
            self.process_game_packets()
        except Exception as err:
            return err

    def runFB(self, ip="127.0.0.1", port=8080):
        if self.system_type == "Linux":
            os.system("chmod +x ./phoenixbuilder")
            con_cmd = rf"./phoenixbuilder -A {self.auth_server} -t fbtoken --no-readline --no-update-check --listen-external {ip}:{port} -c {self.serverNumber} {f'-p {self.serverPassword}' if self.serverPassword else ''}"

        # windows updated "./PRGM" command.
        if self.system_type == "Windows":
            con_cmd = rf".\phoenixbuilder.exe -A {self.auth_server} -t fbtoken --no-readline --no-update-check --listen-external {ip}:{port} -c {self.serverNumber} {f'-p {self.serverPassword}' if self.serverPassword else ''}"
        self.fb_pipe = subprocess.Popen(
            con_cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

    def run_conn(self, ip="127.0.0.1", port=8080, timeout=None):
        connect_fb_start_time = time.time()
        max_con_time = timeout or 10
        while 1:
            try:
                self.con = fbconn.ConnectFB(f"{ip}:{port}")
                Print.print_suc("§a成功连接上FastBuilder.")
                return 1
            except:
                if time.time() - connect_fb_start_time > max_con_time:
                    Print.print_err(f"§4{max_con_time}秒内未连接上FB，已退出")
                    self.close_fb()
                    raise SystemExit
                if self.status == SysStatus.FB_LAUNCH_EXC:
                    Print.print_err("§4连接FB时出现问题，已退出")
                    self.close_fb()
                    raise SystemExit

    def output_fb_msgs_thread(self):
        while 1:
            tmp: str = self.fb_pipe.stdout.readline().decode("utf-8").strip("\n")
            if not tmp:
                continue
            if " 简体中文" in tmp:
                # seems will be unable forever because it's no longer supported.
                try:
                    self.fb_pipe.stdin.write(f"{tmp[1]}\n".encode("utf-8"))
                    self.fb_pipe.stdin.flush()
                    Print.print_inf(f"语言已自动选择为简体中文： [{tmp[1]}]")
                except IndexError:
                    Print.print_war("未能自动选择为简体中文")
            elif "ERROR" in tmp:
                if "Server not found" in tmp:
                    Print.print_err(
                        f"§c租赁服号: {self.serverNumber} 未找到, 有可能是租赁服关闭中, 或是设置了等级或密码"
                    )
                    self.status = SysStatus.NORMAL_EXIT
                elif "Unauthorized rental server number" in tmp:
                    Print.print_err(
                        f"§c租赁服号: {self.serverNumber} ，你还没有该服务器号的卡槽， 请前往用户中心购买"
                    )
                    self.status = SysStatus.NORMAL_EXIT
                elif "Failed to contact with API" in tmp:
                    Print.print_err("§c无法连接到验证服务器, 可能是FB服务器崩溃, 或者是你的IP处于黑名单中")
                    try:
                        Print.print_war("尝试连接到 FastBuilder 验证服务器")
                        requests.get("http://user.fastbuilder.pro", timeout=10)
                        Print.print_err("??? 未知情况， 有可能只是验证服务器崩溃， 用户中心并没有崩溃")
                    except:
                        Print.print_err(
                            "§cFastBuilder服务器无法访问， 请等待修复(加入FastBuilder频道查看详情)"
                        )
                    self.status = SysStatus.NORMAL_EXIT
                elif "Invalid token" in tmp:
                    Print.print_err("§cFastBuilder Token 无法使用， 请重新下载")
                    self.status = SysStatus.NORMAL_EXIT
                elif "netease.report.kick.hint" in tmp:
                    Print.print_err(
                        "§c无法连接到网易租赁服 -> 网易土豆的常见问题，检查你的租赁服状态（等级、是否开启、密码）并重试, 也可能是你的网络问题"
                    )
                    self.status = SysStatus.NORMAL_EXIT
                elif "Press ENTER to exit." in tmp:
                    Print.print_err("§c程序退出")
                    self.status = SysStatus.NORMAL_EXIT
                else:
                    Print.print_with_info(tmp, "§b  FB  §r")

            elif "Transfer: accept new connection @ " in tmp:
                Print.print_with_info(
                    "FastBuilder 监听端口已开放: " + tmp.split()[-1], "§b  FB  "
                )
            elif "Successfully created minecraft dialer." in tmp:
                Print.print_with_info("§e成功创建于 Minecraft 的链接", "§b  FB  §r")
            elif tmp.startswith("panic"):
                Print.print_err(f"FastBuilder 出现问题: {tmp}")
            else:
                Print.print_with_info(tmp, "§b  FB  §r")

    def close_fb(self):
        try:
            self.fb_pipe.stdin.write("exit\n".encode("utf-8"))
            self.fb_pipe.stdin.flush()
        except:
            pass
        try:
            self.fb_pipe.kill()
        except:
            pass
        Print.print_suc("成功关闭FB进程")

    def process_game_packets(self):
        try:
            for packet_bytes in fbconn.RecvGamePacket(self.con):
                packet_type = packet_bytes[0]
                if packet_type not in self.need_listen_packets:
                    continue
                packet_mapping = ujson.loads(
                    fbconn.GamePacketBytesAsIsJsonStr(packet_bytes)
                )
                if packet_type == PacketIDS.CommandOutput:
                    cmd_uuid = packet_mapping["CommandOrigin"]["UUID"].encode()
                    if cmd_uuid in self.cmds_reqs:
                        self.cmds_resp[cmd_uuid] = [time.time(), packet_mapping]
                self.packet_handler(packet_type, packet_mapping)
                if not self.injected and packet_type == PacketIDS.PlayerList:
                    self.injected = True
                    Builtins.createThread(self._launcher_listener)
        except StopIteration:
            pass
        self.status = SysStatus.FB_CRASHED

    def downloadMissingFiles(self):
        "获取缺失文件"
        Print.print_with_info("§d将自动检测缺失文件并补全", "§d 加载 ")
        mirror_src = "https://mirror.ghproxy.com/"
        file_get_src = (
            mirror_src
            + "https://raw.githubusercontent.com/ToolDelta/ToolDelta/main/require_files.json"
        )
        try:
            files_to_get = json.loads(requests.get(file_get_src, timeout=30).text)
        except json.JSONDecodeError:
            Print.print_err("自动下载缺失文件失败: 文件源 JSON 不合法")
            return False
        except requests.Timeout:
            Print.print_err("自动下载缺失文件失败: URL 请求出现问题: 请求超时")
            return False
        except Exception as err:
            Print.print_err(f"自动下载缺失文件失败: URL 请求出现问题: {err}")
            return False
        try:
            Print.print_with_info("§d正在检测需要补全的文件", "§d 加载 ")
            mirrs = files_to_get["Mirror"]
            files = files_to_get[self.system_type]
            for fdir, furl in files.items():
                if not os.path.isfile(fdir):
                    Print.print_inf(f"文件: <{fdir}> 缺失, 正在下载..")
                    succ = False
                    for mirr in mirrs:
                        try:
                            download_file(mirr + "/https://github.com/" + furl, fdir)
                            succ = True
                            break
                        except requests.exceptions.RequestException:
                            Print.print_war("镜像源故障, 正在切换")
                    if not succ:
                        Print.print_err("镜像源全不可用..")
                        return False
                    Print.print_inf(f"文件: <{fdir}> 下载完成        ")
        except requests.Timeout:
            Print.print_err("自动检测文件并补全时出现错误: 超时, 自动跳过")
        except Exception as err:
            Print.print_err(f"自动检测文件并补全时出现错误: {err}")
            return False
        return True

    def init_all_functions(self):
        def sendcmd(cmd: str, waitForResp: bool = False, timeout: int = 30):
            uuid = fbconn.SendMCCommand(self.con, cmd)
            if waitForResp:
                self.cmds_reqs.append(uuid)
                waitStartTime = time.time()
                while 1:
                    res = self.cmds_resp.get(uuid)
                    if res is not None:
                        self.cmds_reqs.remove(uuid)
                        del self.cmds_resp[uuid]
                        return Packet_CommandOutput(res[1])
                    if time.time() - waitStartTime > timeout:
                        self.cmds_reqs.remove(uuid)
                        Print.print_war(f'sendcmd "{cmd}" 超时, 尝试 sendwscmd')
                        return self.sendwscmd(cmd, True, timeout)

        def sendwscmd(cmd: str, waitForResp: bool = False, timeout: int = 30):
            uuid = fbconn.SendWSCommand(self.con, cmd)
            if waitForResp:
                self.cmds_reqs.append(uuid)
                waitStartTime = time.time()
                while 1:
                    res = self.cmds_resp.get(uuid)
                    if res is not None:
                        self.cmds_reqs.remove(uuid)
                        del self.cmds_resp[uuid]
                        return Packet_CommandOutput(res[1])
                    if time.time() - waitStartTime > timeout:
                        self.cmds_reqs.remove(uuid)
                        raise TimeoutError("指令超时")
            else:
                return uuid

        self.sendcmd = sendcmd
        self.sendwscmd = sendwscmd
        self.sendwocmd = staticmethod(
            lambda cmd: fbconn.SendNoResponseCommand(self.con, cmd)
        )
        self.sendPacket = self.sendPacketJson = staticmethod(
            lambda pckID, pck: fbconn.SendGamePacketBytes(
                self.con,
                fbconn.JsonStrAsIsGamePacketBytes(
                    pckID, ujson.dumps(pck, ensure_ascii=False)
                ),
            )
        )
        self.sendfbcmd = staticmethod(lambda cmd: fbconn.SendFBCommand(self.con, cmd))
        self.is_op = None


class FrameNeOmg(StandardFrame):
    # 使用 NeOmega 框架连接到游戏
    launch_type = "NeOmega"

    def __init__(self, serverNumber, password, fbToken, auth_server):
        super().__init__(serverNumber, password, fbToken, auth_server)
        self.status = None
        self.launch_event = threading.Event()
        self.injected = False
        self.omega = None
        self.download_libs()
        self.init_all_functions()
        self.status = SysStatus.LOADING
        self.secret_exit_key = ""

    def set_omega(self, openat_port):
        from .neo_libs import neo_conn

        retries = 0
        while retries <= 10:
            try:
                self.omega = neo_conn.ThreadOmega(
                    connect_type=neo_conn.ConnectType.Remote,
                    address=f"tcp://localhost:{openat_port}",
                    accountOption=neo_conn.AccountOptions(
                        AuthServer=self.auth_server,
                        UserToken=self.fbToken,
                        ServerCode=self.serverNumber,
                        ServerPassword=str(self.serverPassword),
                    ),
                )
                retries = 0
                break
            except Exception as err:
                time.sleep(5)
                retries += 1
                Print.print_war(f"OMEGA 连接失败, 重连: 第 {retries} 次: {err}")
                if retries > 5:
                    Print.print_err("最大重试次数已达到")
                    raise SystemExit

    def start_neomega_proc(self):
        free_port = get_free_port(24016)
        access_point_file = (
            f"neomega_{platform.uname().system.lower()}_access_point_{self.sys_machine}"
        )
        if "TERMUX_VERSION" in os.environ:
            access_point_file = f"neomega_android_access_point_{self.sys_machine}"
        if platform.system() == "Windows":
            access_point_file += ".exe"
        py_file_path = os.path.join(
            os.getcwd(), "tooldelta", "neo_libs", access_point_file
        )
        if platform.uname().system.lower() == "linux":
            os.system(f"chmod +x {shlex.quote(py_file_path)}")
        # 只需要+x即可
        Print.print_inf(f"DEBUG: 将使用端口 {free_port}")
        self.neomg_proc = subprocess.Popen(
            [
                py_file_path,
                "-server",
                self.serverNumber,
                "-T",
                self.fbToken,
                "-access-point-addr",
                f"tcp://localhost:{free_port}",
                "-server-password",
                str(self.serverPassword),
                "-auth-server",
                self.auth_server,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        return free_port

    def msg_show(self):
        def _msg_show_thread():
            str_max_len = 50
            while 1:
                msg_orig = self.neomg_proc.stdout.readline().decode("utf-8").strip("\n")
                if msg_orig in ("", "SIGNAL: exit"):
                    Print.print_with_info("ToolDelta: NEOMG 进程已结束", "§b NOMG ")
                    self.update_status(SysStatus.NORMAL_EXIT)
                    return
                if "[neOmega 接入点]: 就绪" in msg_orig:
                    self.launch_event.set()
                elif f"STATUS CODE: {self.secret_exit_key}" in msg_orig:
                    Print.print_with_info("§a机器人已退出", "§b NOMG ")
                    continue
                while msg_orig:
                    msg = msg_orig[:str_max_len]
                    msg_orig = msg_orig[str_max_len:]
                    Print.print_with_info(msg, "§b NOMG ")

        Builtins.createThread(_msg_show_thread, usage="显示来自NeOmega的信息")

    def update_status(self, new_status):
        self.status = new_status
        if new_status != SysStatus.RUNNING:
            tooldelta.safe_jump()
            self.exit_event.set()  # 设置事件，触发等待结束

    def make_secret_key(self):
        self.secret_exit_key = hex(random.randint(10000, 99999))

    def launch(self):
        self.status = SysStatus.LAUNCHING
        openat_port = self.start_neomega_proc()
        self.msg_show()
        self.launch_event.wait()
        self.make_secret_key()
        self.set_omega(openat_port)
        self.status = SysStatus.RUNNING
        Print.print_suc("已开启 NEOMG 进程")
        pcks = [
            self.omega.get_packet_id_to_name_mapping(i)
            for i in self.need_listen_packets
        ]
        self.omega.listen_packets(pcks, self.packet_handler_parent)
        self._launcher_listener()
        # bug expired
        self.omega.listen_player_chat(lambda _, _2: None)
        Print.print_suc("NEOMEGA 接入已就绪!")
        self.exit_event.wait()  # 等待事件的触发
        if self.status == SysStatus.NORMAL_EXIT:
            return SystemExit("正常退出.")
        if self.status == SysStatus.FB_CRASHED:
            return Exception("NeOmega 已崩溃")
        return SystemError("未知的退出状态")

    def download_libs(self):
        try:
            res = json.loads(
                requests.get(
                    "https://mirror.ghproxy.com/https://raw.githubusercontent.com/ToolDelta/ToolDelta/main/require_files.json"
                ).text
            )
            use_mirror = res["Mirror"][0]
        except Exception as err:
            Print.print_err(f"获取依赖库表出现问题: {err}")
            raise SystemExit
        self.sys_machine = platform.machine().lower()
        if self.sys_machine == "x86_64":
            self.sys_machine = "amd64"
        elif self.sys_machine == "aarch64":
            self.sys_machine = "arm64"
        if "TERMUX_VERSION" in os.environ:
            sys_info_fmt = f"Android:{self.sys_machine.lower()}"
        else:
            sys_info_fmt = f"{platform.uname().system}:{self.sys_machine.lower()}"
        source_dict = res[sys_info_fmt]
        commit_file_path, commit_url = list(res["Commit"].items())[0]
        commit_remote = requests.get(
            use_mirror + "/raw.githubusercontent.com/" + commit_url
        ).text
        commit_local = ""
        commit_file_path = os.path.join(os.getcwd(), commit_file_path)
        if not os.path.isfile(commit_file_path):
            replace_file = True
        else:
            with open(commit_file_path, "r", encoding="utf-8") as f:
                commit_local = f.read()
        replace_file = False
        if commit_local != commit_remote:
            Print.print_war("依赖库版本过期, 将重新下载")
            replace_file = True
        for k, v in source_dict.items():
            pathdir = os.path.join(os.getcwd(), k)
            url = use_mirror + "/https://raw.githubusercontent.com/" + v
            if not os.path.isfile(pathdir) or replace_file:
                Print.print_inf(f"正在下载依赖库 {pathdir} ...")
                try:
                    download_file(url, pathdir)
                except Exception as err:
                    Print.print_err(f"下载依赖库出现问题: {err}")
                    raise SystemExit
        if replace_file:
            with open(
                os.path.join(os.getcwd(), commit_file_path), "w", encoding="utf-8"
            ) as f:
                f.write(commit_remote)
            Print.print_suc("已完成依赖更新！")

    def get_players_and_uuids(self):
        players_uuid = {}
        for i in self.omega.get_all_online_players():
            players_uuid[i.name] = i.uuid
        return players_uuid

    def get_bot_name(self):
        return self.omega.get_bot_name()

    def packet_handler_parent(self, pkt_type, pkt):
        pkt_type = self.omega.get_packet_name_to_id_mapping(pkt_type)
        self.packet_handler(pkt_type, pkt)

    def init_all_functions(self):
        def sendcmd(cmd: str, waitForResp: bool = False, timeout: int = 30):
            if waitForResp:
                res = self.omega.send_player_command_need_response(cmd, timeout)
                if res is None:
                    raise TimeoutError("指令超时")
                return res
            self.omega.send_player_command_omit_response(cmd)
            return

        def sendwscmd(cmd: str, waitForResp: bool = False, timeout: int = 30):
            if waitForResp:
                res = self.omega.send_websocket_command_need_response(cmd, timeout)
                if res is None:
                    raise TimeoutError("指令超时")
                return res
            self.omega.send_websocket_command_omit_response(cmd)
            return

        def sendwocmd(cmd: str):
            self.omega.send_settings_command(cmd)

        def sendPacket(pktID: int, pkt: str):
            self.omega.send_game_packet_in_json_as_is(pktID, pkt)

        def sendfbcmd(_):
            raise AttributeError("NeOmg模式无法发送FBCommand")

        def is_op(player: str):
            return self.omega.get_player_by_name(player).command_permission_level > 2

        self.sendcmd = sendcmd
        self.sendwscmd = sendwscmd
        self.sendwocmd = sendwocmd
        self.sendPacket = self.sendPacketJson = sendPacket
        self.sendfbcmd = sendfbcmd
        self.is_op = is_op


class FrameNeOmgRemote(FrameNeOmg):
    def launch(self):
        try:
            openat_port = int(sys_args_to_dict().get("access-point-port", "24020"))
            if openat_port not in range(65536):
                raise AssertionError
        except (ValueError, AssertionError):
            Print.print_err("启动参数 -access-point-port 错误: 不是1~65535的整数")
        if openat_port == 0:
            Print.print_war("未用启动参数指定链接neOmega接入点开放端口, 尝试使用默认端口 24015")
            Print.print_inf("可使用启动参数 -access-point-port 端口 以指定接入点端口.")
            openat_port = 24015
            return SystemExit
        Print.print_inf(f"将从端口 {openat_port} 连接至 neOmega 接入点.")
        self.set_omega(openat_port)
        Print.print_suc("已连接上 NEOMG 接入点进程.")
        pcks = [
            self.omega.get_packet_id_to_name_mapping(i)
            for i in self.need_listen_packets
        ]
        self.omega.listen_packets(pcks, self.packet_handler_parent)
        self._launcher_listener()
        # bug expired
        self.omega.listen_player_chat(lambda _, _2: None)
        Print.print_suc("NEOMEGA 已就绪")
        self.exit_event.wait()
        return SystemExit()

    def download_libs(self):
        Print.print_inf("以 Remote 启动, 将不会检查库完整性")



import ujson
import os

NoneType = type(None)

PLUGINCFG_DEFAULT = {"配置版本": "0.0.1", "配置项": None}

PLUGINCFG_STANDARD_TYPE = {"配置版本": str, "配置项": [type(None), dict]}


def _CfgIsinstance(obj, typ):
    # 专用于Cfg的类型检测
    if isinstance(typ, type):
        return {
            Cfg.PInt: lambda: isinstance(obj, int) and obj > 0,
            Cfg.NNInt: lambda: isinstance(obj, int) and obj >= 0,
            Cfg.PFloat: lambda: isinstance(obj, float) and obj > 0,
            Cfg.NNFloat: lambda: (isinstance(obj, float) or obj == 0) and obj >= 0,
            Cfg.PNumber: lambda: isinstance(obj, (int, float)) and obj > 0,
            Cfg.NNNumber: lambda: isinstance(obj, (int, float)) and obj >= 0,
        }.get(typ, lambda: isinstance(obj, typ))()
    if isinstance(typ, tuple):
        try:
            for i in typ:
                result = {
                    Cfg.PInt: lambda: isinstance(obj, int) and obj > 0,
                    Cfg.NNInt: lambda: isinstance(obj, int) and obj >= 0,
                    Cfg.PFloat: lambda: isinstance(obj, float) and obj > 0,
                    Cfg.NNFloat: lambda: (isinstance(obj, float) or obj == 0)
                    and obj >= 0,
                    Cfg.PNumber: lambda: isinstance(obj, (int, float)) and obj > 0,
                    Cfg.NNNumber: lambda: isinstance(obj, (int, float)) and obj >= 0,
                }.get(i, lambda: isinstance(obj, typ))()
                if result:
                    return True
            return False
        except TypeError:
            raise ValueError(f"Can't be: {typ}")
    raise ValueError(f"Can't be: {typ}")


def _CfgShowType(typ):
    if type(typ) != type:
        typ = type(typ)
    return {
        Cfg.PInt: "正整数",
        Cfg.NNInt: "非负整数",
        Cfg.PFloat: "正浮点小数",
        Cfg.NNFloat: "非负浮点小数",
        str: "字符串",
        float: "浮点小数",
        int: "整数",
        dict: "json对象",
        list: "列表",
        NoneType: "null",
    }.get(typ, typ.__name__)


FIND_NONE = r"%FindNone"


class Cfg:
    class Group:
        def __init__(self, *keys):
            self.members = keys

        def __repr__(self) -> str:
            return 'Cfg.Group("' + '", "'.join(self.members) + '")'

    class ConfigError(Exception):
        def __init__(self, errStr: str, errPos: list = None):
            if errPos is None:
                errPos = []
            self.errPos = errPos
            self.args = (errStr,)

    class UnneccessaryKey:
        def __init__(self, key):
            self.key = key

        def __repr__(self):
            return f"Cfg.UnneccessaryKey({self.key})"

    class ConfigKeyError(ConfigError):
        "配置json的键错误"

    class ConfigValueError(ConfigError):
        "配置json的值错误"

    class VersionLowError(ConfigError):
        "配置json的版本过低的错误"

    class PInt(int):
        "配置文件的值限制: 正整数"

    class NNInt(int):
        "配置文件的值限制: 非负整数"

    class PFloat(float):
        "配置文件的值限制: 正浮点小数"

    class NNFloat(float):
        "配置文件的值限制: 非负浮点小数"

    class PNumber:
        "配置文件的值限制: 正数"

    class NNNumber:
        "配置文件的值限制: 大于0的数"

    def get_cfg(self, path: str, standard_type: dict):
        # 从path路径获取json文件文本信息, 并按照standard_type给出的标准形式进行检测.
        path = path if path.endswith(".json") else path + ".json"
        with open(path, "r", encoding="utf-8") as f:
            try:
                obj = ujson.load(f)
            except ujson.JSONDecodeError:
                raise self.ConfigValueError("JSON配置文件格式不正确, 请修正或直接删除", None)
        self.check_dict_2(standard_type, obj)
        return obj

    @staticmethod
    def default_cfg(path: str, default: dict, force: bool = False):
        # 向path路径写入json文本, 若文件不存在或参数force为True, 将写入提供的默认json文本
        path = path if path.endswith(".json") else path + ".json"
        if force or not os.path.isfile(path):
            with open(path, "w", encoding="utf-8") as f:
                ujson.dump(default, f, indent=4, ensure_ascii=False)

    @staticmethod
    def exists(path: str):
        return os.path.isfile(path if path.endswith(".json") else path + ".json")

    def getPluginConfigAndVersion(
        self,
        pluginName: str,
        standardType: dict,
        default: dict,
        default_vers: tuple[int, int, int],
    ):
        # 详情见 插件编写指南.md
        assert isinstance(standardType, dict)
        p = "插件配置文件/" + pluginName
        if not self.exists(p) and default:
            defaultCfg = PLUGINCFG_DEFAULT.copy()
            defaultCfg["配置项"] = default
            defaultCfg["配置版本"] = ".".join([str(n) for n in default_vers])
            self.check_auto(standardType, default)
            self.default_cfg(p + ".json", defaultCfg, force=True)
        cfg_stdtyp = PLUGINCFG_STANDARD_TYPE.copy()
        cfg_stdtyp["配置项"] = standardType
        cfgGet = self.get_cfg(p, cfg_stdtyp)
        cfgVers = tuple(int(c) for c in cfgGet["配置版本"].split("."))
        return cfgGet["配置项"], cfgVers

    def check_auto(self, standard, val, fromkey="?"):
        if fromkey == FIND_NONE:
            raise ValueError("不允许传入FindNone")
        if isinstance(standard, type):
            if not _CfgIsinstance(val, standard):
                raise self.ConfigValueError(
                    f'JSON键"{fromkey}" 对应值的类型不正确: 需要 {_CfgShowType(standard)}, 实际上为 {_CfgShowType(val)}'
                )
        elif isinstance(standard, list):
            # [%list] or [multi[type/dict]]
            self.check_list_2(standard, val)
        elif isinstance(standard, dict):
            self.check_dict_2(standard, val)
        else:
            raise ValueError(
                f'JSON键 "{fromkey}"未曾遇到过的类型: {standard.__class__.__name__}, 另外两个参数 standard={standard}, val={val}'
            )

    def check_dict_2(self, pattern: dict, jsondict: dict):
        """
        按照给定的标准配置样式比对传入的配置文件jsondict, 对不上则引发相应异常

        参数:
            pattern: 标准样式dict
            jsondict: 待检测的配置文件dict
        """
        for key, std_val in pattern.items():
            if key == r"%any":
                # ANY key
                for js_key, js_val in jsondict.items():
                    self.check_auto(std_val, js_val, js_key)
            elif isinstance(key, self.Group):
                # any key in members?
                js_val = FIND_NONE
                for member_key in key.members:
                    val_get = jsondict.get(member_key, FIND_NONE)
                    if val_get != FIND_NONE:
                        self.check_auto(std_val, val_get, member_key)
            elif isinstance(key, self.UnneccessaryKey):
                val_get = jsondict.get(key.key, FIND_NONE)
                if val_get != FIND_NONE:
                    self.check_auto(std_val, val_get, key.key)
            else:
                val_get = jsondict.get(key, FIND_NONE)
                if val_get == FIND_NONE:
                    raise self.ConfigKeyError(f"不存在的JSON键: {key}")
                self.check_auto(std_val, val_get, key)

    def check_list_2(self, pattern: list, value, fromkey="?"):
        if not isinstance(pattern, list):
            raise ValueError("不是合法的标准列表检测样式")
        if len(pattern) == 0:
            raise ValueError("标准检测列表的长度不能为0")
        if isinstance(pattern[0], str) and pattern[0].startswith(r"%list"):
            if not isinstance(value, list):
                raise self.ConfigValueError(
                    f'JSON键 "{fromkey}" 需要列表 而不是 {_CfgShowType(value)}'
                )
            limitNumber = 0
            if pattern[0].replace(r"%list", "").isnumeric():
                limitNumber = int(pattern[0].replace(r"%list", ""))
            if limitNumber and len(value) != limitNumber:
                raise self.ConfigValueError(
                    f'JSON键 "{fromkey}" 所对应的值列表有误: 需要 {limitNumber} 项, 实际上为 {len(value)} 项'
                )
            for val in value:
                self.check_auto(pattern[1], val, fromkey)
        else:
            for single_type in pattern:
                try:
                    self.check_auto(single_type, value, fromkey)
                    return
                except:
                    pass
            raise self.ConfigValueError(
                f"JSON列表的值 \"{fromkey}\" 类型不正确: 需要 {' 或 '.join(_CfgShowType(i) for i in pattern)}, 实际上为 {_CfgShowType(value)}"
            )

    checkDict = check_dict_2


if __name__ == "__main__":
    # Test Part
    try:
        test_cfg = 12
        a_std = [int, [r"%list", str]]
        Cfg().check_auto(a_std, test_cfg)
    except Cfg.ConfigError:
        import traceback

        print(traceback.format_exc())
